import ctypes
from ctypes import c_uint
from ctypes.wintypes import HWND, LPARAM

import pytest
from System.Windows.Forms import TextBox

from .base import SimpleProbe
from .properties import toga_x_text_align


class TextInputProbe(SimpleProbe):
    native_class = TextBox
    fixed_height = 23

    @property
    def value(self):
        return self._placeholder if self.placeholder_visible else self.native.Text

    @property
    def value_hidden(self):
        return self.native.UseSystemPasswordChar

    @property
    def _placeholder(self):
        buffer = ctypes.create_unicode_buffer(1024)
        result = ctypes.windll.user32.SendMessageW(
            HWND(self.native.Handle.ToInt32()),
            c_uint(0x1502),  # EM_GETCUEBANNER
            buffer,
            LPARAM(ctypes.sizeof(buffer)),
        )
        if not result:
            raise RuntimeError("EM_GETCUEBANNER failed")
        return buffer.value

    @property
    def placeholder_visible(self):
        return not self.native.Text

    @property
    def placeholder_hides_on_focus(self):
        False

    @property
    def readonly(self):
        return self.native.ReadOnly

    @property
    def text_align(self):
        return toga_x_text_align(self.native.TextAlign)

    def assert_vertical_text_align(self, expected):
        # Vertical text alignment isn't configurable in this native widget.
        pass

    def set_cursor_at_end(self):
        pytest.skip("Cursor positioning not supported on this platform")



from System.Windows.Forms import (
    BorderStyle,
    Orientation,
    SplitContainer as NativeSplitContainer,
)

from toga.constants import Direction

from ..container import Container
from ..libs.wrapper import WeakrefCallable
from .base import Widget


class SplitContainer(Widget):
    def create(self):
        self.native = NativeSplitContainer()
        self.native.SplitterMoved += WeakrefCallable(self.winforms_splitter_moved)

        # Despite what the BorderStyle documentation says, there is no border by default
        # (at least on Windows 10), which would make the split bar invisible.
        self.native.BorderStyle = BorderStyle.Fixed3D

        self.panels = (Container(self.native.Panel1), Container(self.native.Panel2))
        self.pending_position = None

    def winforms_splitter_moved(self, sender, event):
        self.resize_content()

    def set_bounds(self, x, y, width, height):
        super().set_bounds(x, y, width, height)

        force_refresh = False
        if self.pending_position:
            self.set_position(self.pending_position)
            self.pending_position = None
            force_refresh = True  # Content has changed

        self.resize_content(force_refresh=force_refresh)

    def set_content(self, content, flex):
        # In case content moves from one panel to another, make sure it's removed first
        # so it doesn't get removed again by set_content.
        for panel in self.panels:
            panel.clear_content()

        for panel, widget in zip(self.panels, content):
            panel.set_content(widget)

        self.pending_position = flex[0] / sum(flex)

    def get_direction(self):
        return {
            Orientation.Vertical: Direction.VERTICAL,
            Orientation.Horizontal: Direction.HORIZONTAL,
        }[self.native.Orientation]

    def set_direction(self, value):
        position = self.get_position()

        self.native.Orientation = {
            Direction.VERTICAL: Orientation.Vertical,
            Direction.HORIZONTAL: Orientation.Horizontal,
        }[value]

        self.set_position(position)

    def get_position(self):
        return self.native.SplitterDistance / self.get_max_position()

    def set_position(self, position):
        self.native.SplitterDistance = round(position * self.get_max_position())

    def get_max_position(self):
        return (
            self.native.Width
            if self.get_direction() == Direction.VERTICAL
            else self.native.Height
        )

    def resize_content(self, **kwargs):
        for panel in self.panels:
            size = panel.native_parent.ClientSize
            panel.resize_content(size.Width, size.Height, **kwargs)



from System.Drawing import Color
from travertino.colors import TRANSPARENT, rgba

CACHE = {TRANSPARENT: Color.Transparent}


def native_color(c):
    try:
        color = CACHE[c]
    except KeyError:
        color = Color.FromArgb(
            int(c.rgba.a * 255),
            int(c.rgba.r),
            int(c.rgba.g),
            int(c.rgba.b),
        )
        CACHE[c] = color

    return color


def toga_color(c):
    return rgba(c.R, c.G, c.B, c.A / 255)



from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from travertino.size import at_least

from toga.colors import TRANSPARENT

from ..libs.wrapper import WeakrefCallable
from .base import Widget


class Button(Widget):

    def create(self):
        self.native = WinForms.Button()
        self.native.AutoSizeMode = WinForms.AutoSizeMode.GrowAndShrink
        self.native.Click += WeakrefCallable(self.winforms_click)

        self._icon = None

    def winforms_click(self, sender, event):
        self.interface.on_press()

    def get_text(self):
        value = self.native.Text
        # Normalize a standalone ZERO WIDTH SPACE to an empty string.
        if value == "\u200b":
            return ""
        return value

    def set_text(self, text):
        if text == "":
            # An empty label would cause the widget's height to collapse, so display a
            # Unicode ZERO WIDTH SPACE instead.
            text = "\u200b"
        self.native.Text = text

    def get_icon(self):
        return self._icon

    def set_icon(self, icon):
        self._icon = icon
        if icon:
            self.native.Image = icon._impl.native.ToBitmap()
        else:
            self.native.Image = None

    def set_background_color(self, color):
        super().set_background_color(
            self._default_background_color if color in {None, TRANSPARENT} else color
        )

    def rehint(self):
        self.interface.intrinsic.width = self.scale_out(
            at_least(self.native.PreferredSize.Width), ROUND_UP
        )
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



from System.Windows.Forms import Screen as WinFormsScreen

from toga.images import Image as TogaImage

from .probe import BaseProbe


class ScreenProbe(BaseProbe):
    def __init__(self, screen):
        super().__init__()
        self.screen = screen
        self._impl = screen._impl
        self.native = screen._impl.native
        assert isinstance(self.native, WinFormsScreen)

    def get_screenshot(self, format=TogaImage):
        return self.screen.as_image(format=format)



from toga.sources import Row

from .table import Table


# Wrap a DetailedList source to make it compatible with a Table.
class TableSource:
    def __init__(self, interface):
        self.interface = interface

    def __len__(self):
        return len(self.interface.data)

    def __getitem__(self, index):
        row = self.interface.data[index]
        title, subtitle, icon = (
            getattr(row, attr, None) for attr in self.interface.accessors
        )
        return Row(title=(icon, title), subtitle=subtitle)


class DetailedList(Table):
    # The following methods are overridden from Table.
    @property
    def _headings(self):
        return None

    @property
    def _accessors(self):
        return ("title", "subtitle")

    @property
    def _multiple_select(self):
        return False

    @property
    def _data(self):
        return self._table_source

    def create(self):
        super().create()
        self._table_source = TableSource(self.interface)

    def add_action_events(self):
        # DetailedList doesn't have an on_activate_handler.
        pass

    def set_primary_action_enabled(self, enabled):
        self.primary_action_enabled = enabled

    def set_secondary_action_enabled(self, enabled):
        self.secondary_action_enabled = enabled

    def set_refresh_enabled(self, enabled):
        self.refresh_enabled = enabled

    after_on_refresh = None



import pytest
from pytest import approx
from System import EventArgs, Object
from System.Drawing import SystemColors
from System.Windows.Forms import MouseButtons, MouseEventArgs

from toga.style.pack import JUSTIFY, LEFT

from ..probe import BaseProbe
from .properties import toga_color


class SimpleProbe(BaseProbe):
    def __init__(self, widget):
        self.app = widget.app
        self.widget = widget
        self.impl = widget._impl
        super().__init__(self.impl.native)
        assert isinstance(self.native, self.native_class)

    def assert_container(self, container):
        assert self.widget._impl.container is container._impl.container
        assert self.native.Parent is not None
        assert Object.ReferenceEquals(
            self.native.Parent,
            container._impl.container.native_content,
        )

    def assert_not_contained(self):
        assert self.widget._impl.container is None
        assert self.native.Parent is None

    def assert_text_align(self, expected):
        # Winforms doesn't have a "Justified" text alignment; it falls back to LEFT
        actual = self.text_align
        if expected == JUSTIFY:
            assert actual == LEFT
        else:
            assert actual == expected

    @property
    def enabled(self):
        return self.native.Enabled

    @property
    def color(self):
        if self.native.ForeColor == SystemColors.WindowText:
            return None
        else:
            return toga_color(self.native.ForeColor)

    @property
    def background_color(self):
        return (
            toga_color(self.native.BackColor),
            toga_color(self.widget.parent._impl.native.BackColor),
            (
                # self.impl.interface.style.background_color can be None or TRANSPARENT
                # and so there will be no alpha value on them. In such cases return 0
                # as the original alpha value.
                getattr(self.widget.style.background_color, "a", 0)
            ),
        )

    @property
    def hidden(self):
        return not self.native.Visible

    @property
    def shrink_on_resize(self):
        return True

    def assert_layout(self, size, position):
        # Widget is contained and in a window.
        assert self.widget._impl.container is not None
        assert self.native.Parent is not None

        # size and position is as expected.
        assert (self.width, self.height) == approx(size, abs=1)
        assert (self.x, self.y) == approx(position, abs=1)

    async def press(self):
        self.native.OnClick(EventArgs.Empty)

    def mouse_event(self, x=0, y=0, **kwargs):
        kwargs = {**dict(button=MouseButtons.Left, clicks=1, delta=0), **kwargs}
        return MouseEventArgs(
            x=round(x * self.scale_factor), y=round(y * self.scale_factor), **kwargs
        )

    @property
    def is_hidden(self):
        return not self.native.Visible

    @property
    def has_focus(self):
        return self.native.ContainsFocus

    async def undo(self):
        pytest.skip("Undo not supported on this platform")

    async def redo(self):
        pytest.skip("Redo not supported on this platform")



import System.Windows.Forms

from .base import SimpleProbe


class ProgressBarProbe(SimpleProbe):
    native_class = System.Windows.Forms.ProgressBar

    @property
    def is_determinate(self):
        # Winforms doesn't have an explicit flag for describing determinate
        # progress bars; use the shadow flag.
        return self.widget._impl._determinate

    @property
    def is_animating_indeterminate(self):
        # We can use the winforms style to identify animating indeterminates
        return self.native.Style == System.Windows.Forms.ProgressBarStyle.Marquee

    @property
    def position(self):
        return self.native.Value / self.native.Maximum

    async def wait_for_animation(self):
        # WinForms ProgressBar has internal animation handling;
        #  no special handling required.
        pass



import pytest
import System.Windows.Forms

from .base import SimpleProbe


class ButtonProbe(SimpleProbe):
    native_class = System.Windows.Forms.Button

    @property
    def text(self):
        # Normalize the zero width space to the empty string.
        if self.native.Text == "\u200b":
            return ""
        return self.native.Text

    def assert_no_icon(self):
        assert self.native.Image is None

    def assert_icon_size(self):
        icon = self.native.Image
        if icon:
            assert (icon.Size.Width, icon.Size.Height) == (32, 32)
        else:
            pytest.fail("Icon does not exist")



import weakref


class WeakrefCallable:
    """
    A wrapper for callable that holds a weak reference to it.

    This can be useful in particular when setting winforms event handlers, to avoid
    cyclical reference cycles between Python and the .NET CLR that are detected neither
    by the Python garbage collector nor the C# garbage collector.
    """

    def __init__(self, function):
        try:
            self.ref = weakref.WeakMethod(function)
        except TypeError:  # pragma: no cover
            self.ref = weakref.ref(function)

    def __call__(self, *args, **kwargs):
        function = self.ref()
        if function:  # pragma: no branch
            return function(*args, **kwargs)



from ctypes import HRESULT, POINTER, windll, wintypes

GetScaleFactorForMonitor = windll.shcore.GetScaleFactorForMonitor
GetScaleFactorForMonitor.restype = HRESULT
GetScaleFactorForMonitor.argtypes = [wintypes.HMONITOR, POINTER(wintypes.UINT)]



from ctypes import wintypes

from System.Drawing import (
    Bitmap,
    Graphics,
    Imaging,
    Point,
    Size as WinSize,
)
from System.IO import MemoryStream

from toga import App
from toga.screens import Screen as ScreenInterface
from toga.types import Position, Size

from .libs import shcore, user32
from .widgets.base import Scalable


class Screen(Scalable):
    _instances = {}

    def __new__(cls, native):
        if native in cls._instances:
            return cls._instances[native]
        else:
            instance = super().__new__(cls)
            instance.interface = ScreenInterface(_impl=instance)
            instance.native = native
            cls._instances[native] = instance
            return instance

    @property
    def dpi_scale(self):
        screen_rect = wintypes.RECT(
            self.native.Bounds.Left,
            self.native.Bounds.Top,
            self.native.Bounds.Right,
            self.native.Bounds.Bottom,
        )
        hMonitor = user32.MonitorFromRect(screen_rect, user32.MONITOR_DEFAULTTONEAREST)
        pScale = wintypes.UINT()
        shcore.GetScaleFactorForMonitor(hMonitor, pScale)
        return pScale.value / 100

    def get_name(self):
        name = self.native.DeviceName
        # WinForms Display naming convention is "\\.\DISPLAY1". Remove the
        # non-text part to prevent any errors due to non-escaped characters.
        return name.split("\\")[-1]

    # Screen.origin is scaled according to the DPI of the primary screen, because there
    # is no better choice that could cover screens of multiple DPIs.
    def get_origin(self) -> Position:
        primary_screen = App.app._impl.get_primary_screen()
        bounds = self.native.Bounds
        return Position(
            primary_screen.scale_out(bounds.X), primary_screen.scale_out(bounds.Y)
        )

    # Screen.size is scaled according to the screen's own DPI, to be consistent with the
    # scaling of Window size and content.
    def get_size(self) -> Size:
        bounds = self.native.Bounds
        return Size(self.scale_out(bounds.Width), self.scale_out(bounds.Height))

    def get_image_data(self):
        bitmap = Bitmap(*self.get_size())
        graphics = Graphics.FromImage(bitmap)
        source_point = Point(*self.get_origin())
        destination_point = Point(0, 0)
        copy_size = WinSize(*self.get_size())
        graphics.CopyFromScreen(source_point, destination_point, copy_size)
        stream = MemoryStream()
        bitmap.Save(stream, Imaging.ImageFormat.Png)
        return bytes(stream.ToArray())



import asyncio
from unittest.mock import Mock

from System import Array as WinArray, String as WinString


class DialogsMixin:
    supports_multiple_select_folder = False

    def _setup_dialog_result(self, dialog, char, alt=False, pre_close_test_method=None):
        # Install an overridden show method that invokes the original,
        # but then closes the open dialog.
        orig_show = dialog._impl.show

        def automated_show(host_window, future):
            orig_show(host_window, future)

            async def _close_dialog():
                # Give the inner event loop a chance to start. The MessageBox dialogs
                # work with sleep(0), but the file dialogs require it to be positive
                # for some reason.
                await asyncio.sleep(0.001)

                try:
                    if pre_close_test_method:
                        pre_close_test_method(dialog)
                finally:
                    try:
                        await self.type_character(char, alt=alt)
                    except Exception as e:
                        # An error occurred closing the dialog; that means the dialog
                        # isn't what as expected, so record that in the future.
                        future.set_exception(e)

            asyncio.create_task(_close_dialog(), name="close-dialog")

        dialog._impl.show = automated_show

    def setup_info_dialog_result(self, dialog, pre_close_test_method=None):
        self._setup_dialog_result(
            dialog, "\n", pre_close_test_method=pre_close_test_method
        )

    def setup_question_dialog_result(self, dialog, result):
        self._setup_dialog_result(dialog, "y" if result else "n")

    def setup_confirm_dialog_result(self, dialog, result):
        self._setup_dialog_result(dialog, "\n" if result else "<esc>")

    def setup_error_dialog_result(self, dialog):
        self._setup_dialog_result(dialog, "\n")

    def setup_stack_trace_dialog_result(self, dialog, result):
        self._setup_dialog_result(
            dialog,
            {None: "o", True: "r", False: "q"}[result],
            alt=True,
        )

    def setup_save_file_dialog_result(self, dialog, result):
        self._setup_dialog_result(dialog, "\n" if result else "<esc>")

    def setup_open_file_dialog_result(self, dialog, result, multiple_select):
        if result is None:
            self._setup_dialog_result(dialog, "<esc>")
        else:
            if multiple_select:
                # Since we are mocking selected_path(), it's never actually invoked
                # under test conditions. Call it just to confirm that it returns the
                # type we think it does.
                assert isinstance(
                    dialog._impl.selected_paths(),
                    type(WinArray.CreateInstance(WinString, 0)),
                )

                # native.FileNames is read-only, so we have to mock the mechanism
                # returning the result
                dialog._impl.native.FileName = str(result[0])  # Enable the OK button
                dialog._impl.selected_paths = Mock(
                    return_value=[str(path) for path in result]
                )
            else:
                dialog._impl.native.FileName = str(result)

            self._setup_dialog_result(dialog, "\n")

    def setup_select_folder_dialog_result(self, dialog, result, multiple_select):
        if result is None:
            self._setup_dialog_result(dialog, "<esc>")
        else:
            dialog._impl.native.SelectedPath = str(
                result[-1] if multiple_select else result
            )
            self._setup_dialog_result(dialog, "\n")

    def is_modal_dialog(self, dialog):
        return True



import asyncio

import pytest
from Microsoft.Web.WebView2.WinForms import WebView2

from .base import SimpleProbe


def region_eq(r1, r2):
    return (
        pytest.approx(r1[0]["lat"]) == pytest.approx(r2[0]["lat"])
        and pytest.approx(r1[1]["lat"]) == pytest.approx(r2[1]["lat"])
        and pytest.approx(r1[0]["lng"]) == pytest.approx(r2[0]["lng"])
        and pytest.approx(r1[1]["lng"]) == pytest.approx(r2[1]["lng"])
    )


class MapViewProbe(SimpleProbe):
    native_class = WebView2

    async def _map_region(self):
        northeast = self.impl._invoke("map.getBounds().getNorthEast();")
        southwest = self.impl._invoke("map.getBounds().getSouthWest();")
        return northeast, southwest

    async def tile_longitude_span(self):
        northeast, southwest = await self._map_region()
        return 256 * (northeast["lng"] - southwest["lng"]) / self.width

    @property
    def pin_count(self):
        return int(self.impl._invoke("Object.keys(pins).length;"))

    async def select_pin(self, pin):
        pytest.skip("Winforms MapView doesn't support selecting pins")

    async def wait_for_map(self, message, max_delay=0.5):
        initial = await self._map_region()
        previous = initial
        panning = True

        # Iterate until 2 successive reads of the region, 0.2s apart, return the same
        # value; or we've been waiting max_delay seconds. If confirm_pan is True, also
        # confirm that the value has actually changed from the initial value.
        tick_count = 0
        delta = 0.2
        while panning and tick_count < (max_delay / delta):
            await asyncio.sleep(delta)
            current = await self._map_region()

            # If the region isn't stable, we're still panning.
            panning = not region_eq(current, previous)

            # If the region is the same as the original region,
            # we're still waiting for panning to start. It's possible
            # panning has finished, but it's better to timeout on the
            # tick count to be sure
            if region_eq(current, initial):
                panning = True

            previous = current
            tick_count += 1

        await self.redraw(message)



from math import degrees

import System.Windows.Forms as WinForms
from System import Array
from System.Drawing import (
    Bitmap,
    Graphics,
    Pen,
    PointF,
    Rectangle,
    RectangleF,
    SolidBrush,
    StringFormat,
)
from System.Drawing.Drawing2D import (
    FillMode,
    GraphicsPath,
    Matrix,
    PixelOffsetMode,
    SmoothingMode,
)
from System.Drawing.Imaging import ImageFormat
from System.IO import MemoryStream

from toga.colors import TRANSPARENT
from toga.constants import Baseline, FillRule
from toga.widgets.canvas import arc_to_bezier, sweepangle
from toga_winforms.colors import native_color

from ..libs.wrapper import WeakrefCallable
from .box import Box


class WinformContext:
    def __init__(self):
        super().__init__()
        self.graphics = None
        self.matrix = Matrix()
        self.clear_paths()

    def clear_paths(self):
        self.paths = []
        self.add_path()

    @property
    def current_path(self):
        return self.paths[-1]

    def add_path(self, start_point=None):
        self.paths.append(GraphicsPath())
        self.start_point = start_point

    # Because the GraphicsPath API works in terms of segments rather than points, it has
    # no equivalent to move_to, and we must save that point manually. In all other
    # situations, we can get the last point from the GraphicsPath itself.
    #
    # default_x and default_y should be set as described in the HTML spec under "ensure
    # there is a subpath".
    def get_last_point(self, default_x, default_y):
        if self.current_path.PointCount:
            return self.current_path.GetLastPoint()
        elif self.start_point:
            return self.start_point
        else:
            return PointF(default_x, default_y)

    def print_path(self, path=None):  # pragma: no cover
        if path is None:
            path = self.current_path
        print(
            "\n".join(
                str((ptype, point.X, point.Y))
                for ptype, point in zip(path.PathTypes, path.PathPoints)
            )
        )


class Canvas(Box):
    def create(self):
        super().create()
        self._default_background_color = TRANSPARENT
        self.native.DoubleBuffered = True
        self.native.Paint += WeakrefCallable(self.winforms_paint)
        self.native.Resize += WeakrefCallable(self.winforms_resize)
        self.native.MouseDown += WeakrefCallable(self.winforms_mouse_down)
        self.native.MouseMove += WeakrefCallable(self.winforms_mouse_move)
        self.native.MouseUp += WeakrefCallable(self.winforms_mouse_up)
        self.string_format = StringFormat.GenericTypographic
        self.dragging = False
        self.states = []

    # The control automatically paints the background color, so painting it again here
    # would give incorrect results if it was semi-transparent. But we do paint it in
    # get_image_data.
    def winforms_paint(self, panel, event, *args):
        context = WinformContext()
        self.reset_transform(context)
        context.graphics = event.Graphics
        context.graphics.PixelOffsetMode = PixelOffsetMode.HighQuality
        context.graphics.SmoothingMode = SmoothingMode.AntiAlias
        self.interface.context._draw(self, draw_context=context)

    def winforms_resize(self, *args):
        self.interface.on_resize(
            width=self.scale_out(self.native.Width),
            height=self.scale_out(self.native.Height),
        )

    def winforms_mouse_down(self, obj, mouse_event):
        x, y = map(self.scale_out, (mouse_event.X, mouse_event.Y))
        if mouse_event.Button == WinForms.MouseButtons.Left:
            if mouse_event.Clicks == 2:
                self.interface.on_activate(x, y)
            else:
                self.interface.on_press(x, y)
                self.dragging = True
        elif mouse_event.Button == WinForms.MouseButtons.Right:
            self.interface.on_alt_press(x, y)
            self.dragging = True
        else:  # pragma: no cover
            pass

    def winforms_mouse_move(self, obj, mouse_event):
        if not self.dragging:
            return
        x, y = map(self.scale_out, (mouse_event.X, mouse_event.Y))
        if mouse_event.Button == WinForms.MouseButtons.Left:
            self.interface.on_drag(x, y)
        elif mouse_event.Button == WinForms.MouseButtons.Right:
            self.interface.on_alt_drag(x, y)
        else:  # pragma: no cover
            pass

    def winforms_mouse_up(self, obj, mouse_event):
        self.dragging = False
        x, y = map(self.scale_out, (mouse_event.X, mouse_event.Y))
        if mouse_event.Button == WinForms.MouseButtons.Left:
            self.interface.on_release(x, y)
        elif mouse_event.Button == WinForms.MouseButtons.Right:
            self.interface.on_alt_release(x, y)
        else:  # pragma: no cover
            pass

    def redraw(self):
        self.native.Invalidate()

    # Context management

    def push_context(self, draw_context, **kwargs):
        self.states.append(draw_context.matrix)
        draw_context.matrix = draw_context.matrix.Clone()

    def pop_context(self, draw_context, **kwargs):
        draw_context.matrix = self.states.pop()

    # Basic paths

    def begin_path(self, draw_context, **kwargs):
        draw_context.clear_paths()

    # We don't use current_path.CloseFigure, because that causes the dash pattern to
    # start on the last segment of the path rather than the first one.
    def close_path(self, draw_context, **kwargs):
        if draw_context.current_path.PointCount:
            start = draw_context.current_path.PathPoints[0]
            draw_context.current_path.AddLine(
                draw_context.current_path.GetLastPoint(), start
            )
            self.move_to(start.X, start.Y, draw_context)

    def move_to(self, x, y, draw_context, **kwargs):
        draw_context.add_path(PointF(x, y))

    def line_to(self, x, y, draw_context, **kwargs):
        draw_context.current_path.AddLine(
            draw_context.get_last_point(x, y), PointF(x, y)
        )

    # Basic shapes

    def bezier_curve_to(self, cp1x, cp1y, cp2x, cp2y, x, y, draw_context, **kwargs):
        draw_context.current_path.AddBezier(
            draw_context.get_last_point(cp1x, cp1y),
            PointF(cp1x, cp1y),
            PointF(cp2x, cp2y),
            PointF(x, y),
        )

    # A Quadratic curve is a dimensionally reduced Bézier Cubic curve;
    # we can convert the single Quadratic control point into the
    # 2 control points required for the cubic Bézier.
    def quadratic_curve_to(self, cpx, cpy, x, y, draw_context, **kwargs):
        last_point = draw_context.get_last_point(cpx, cpy)
        x0, y0 = (last_point.X, last_point.Y)
        draw_context.current_path.AddBezier(
            last_point,
            PointF(
                x0 + 2 / 3 * (cpx - x0),
                y0 + 2 / 3 * (cpy - y0),
            ),
            PointF(
                x + 2 / 3 * (cpx - x),
                y + 2 / 3 * (cpy - y),
            ),
            PointF(x, y),
        )

    def arc(
        self,
        x,
        y,
        radius,
        startangle,
        endangle,
        counterclockwise,
        draw_context,
        **kwargs,
    ):
        self.ellipse(
            x,
            y,
            radius,
            radius,
            0,
            startangle,
            endangle,
            counterclockwise,
            draw_context,
            **kwargs,
        )

    def ellipse(
        self,
        x,
        y,
        radiusx,
        radiusy,
        rotation,
        startangle,
        endangle,
        counterclockwise,
        draw_context,
        **kwargs,
    ):
        matrix = Matrix()
        matrix.Translate(x, y)
        matrix.Rotate(degrees(rotation))
        matrix.Scale(radiusx, radiusy)
        matrix.Rotate(degrees(startangle))

        points = Array[PointF](
            [
                PointF(x, y)
                for x, y in arc_to_bezier(
                    sweepangle(startangle, endangle, counterclockwise)
                )
            ]
        )
        matrix.TransformPoints(points)

        start = draw_context.start_point
        if start and not draw_context.current_path.PointCount:
            draw_context.current_path.AddLine(start, start)
        draw_context.current_path.AddBeziers(points)

    def rect(self, x, y, width, height, draw_context, **kwargs):
        draw_context.add_path()
        rect = RectangleF(x, y, width, height)
        draw_context.current_path.AddRectangle(rect)
        draw_context.add_path()

    # Drawing Paths

    def fill(self, color, fill_rule, draw_context, **kwargs):
        brush = SolidBrush(native_color(color))
        for path in draw_context.paths:
            if fill_rule == FillRule.EVENODD:
                path.FillMode = FillMode.Alternate
            else:  # Default to NONZERO
                path.FillMode = FillMode.Winding
            path.Transform(draw_context.matrix)
            draw_context.graphics.FillPath(brush, path)
        draw_context.clear_paths()

    def stroke(self, color, line_width, line_dash, draw_context, **kwargs):
        pen = Pen(native_color(color), self.scale_in(line_width, rounding=None))
        if line_dash is not None:
            pen.DashPattern = [ld / line_width for ld in line_dash]

        for path in draw_context.paths:
            path.Transform(draw_context.matrix)
            draw_context.graphics.DrawPath(pen, path)
        draw_context.clear_paths()

    # Transformations

    def rotate(self, radians, draw_context, **kwargs):
        draw_context.matrix.Rotate(degrees(radians))

    def scale(self, sx, sy, draw_context, **kwargs):
        draw_context.matrix.Scale(sx, sy)

    def translate(self, tx, ty, draw_context, **kwargs):
        draw_context.matrix.Translate(tx, ty)

    def reset_transform(self, draw_context, **kwargs):
        draw_context.matrix.Reset()
        self.scale(self.dpi_scale, self.dpi_scale, draw_context)

    # Text
    def _line_height(self, font, line_height):
        if line_height is None:
            return font.metric("LineSpacing")
        else:
            # Get size in CSS pixels
            return (font.native.SizeInPoints * 96 / 72) * line_height

    def write_text(
        self, text, x, y, font, baseline, line_height, draw_context, **kwargs
    ):
        for op in ["fill", "stroke"]:
            if color := kwargs.pop(f"{op}_color", None):
                self._text_path(text, x, y, font, baseline, line_height, draw_context)
                getattr(self, op)(color, draw_context=draw_context, **kwargs)

    def _text_path(self, text, x, y, font, baseline, line_height, draw_context):
        lines = text.splitlines()
        scaled_line_height = self._line_height(font, line_height)
        total_height = scaled_line_height * len(lines)

        if baseline == Baseline.TOP:
            top = y
        elif baseline == Baseline.MIDDLE:
            top = y - (total_height / 2)
        elif baseline == Baseline.BOTTOM:
            top = y - total_height
        else:
            # Default to Baseline.ALPHABETIC
            top = y - font.metric("CellAscent")

        for line_num, line in enumerate(lines):
            draw_context.current_path.AddString(
                line,
                font.native.FontFamily,
                font.native.Style.value__,
                font.metric("EmHeight"),
                PointF(x, top + (scaled_line_height * line_num)),
                self.string_format,
            )

    def measure_text(self, text, font, line_height):
        graphics = self.native.CreateGraphics()
        sizes = [
            graphics.MeasureString(line, font.native, 2**31 - 1, self.string_format)
            for line in text.splitlines()
        ]
        return (
            self.scale_out(max(size.Width for size in sizes)),
            self._line_height(font, line_height) * len(sizes),
        )

    def get_image_data(self):
        # Winforms backgrounds don't honor transparency, so the background that is
        # rendered to screen manually computes the blended color. However, we want the
        # image to reflect the background color that has actually been applied to the
        # image. Temporarily switch out the manually alpha blended background color to
        # the native system produced background color. Suspending the layout means this
        # change isn't visible to the user.
        self.native.SuspendLayout()
        current_background_color = self.interface.style.background_color
        self.native.BackColor = native_color(current_background_color)

        width, height = (self.native.Width, self.native.Height)
        bitmap = Bitmap(width, height)
        rect = Rectangle(0, 0, width, height)
        graphics = Graphics.FromImage(bitmap)
        graphics.Clear(self.native.BackColor)
        self.native.OnPaint(WinForms.PaintEventArgs(graphics, rect))

        stream = MemoryStream()
        bitmap.Save(stream, ImageFormat.Png)

        # Switch back to the manually alpha blended background color, and resume layout
        # updates.
        self.set_background_color(current_background_color)
        self.native.ResumeLayout()

        return bytes(stream.ToArray())



import asyncio
from ctypes import byref, c_void_p, windll, wintypes

from pytest import approx
from System.Windows.Forms import Screen, SendKeys

import toga

from .fonts import FontMixin

KEY_CODES = {
    f"<{name}>": f"{{{name.upper()}}}"
    for name in ["esc", "up", "down", "left", "right", "home", "end"]
}
KEY_CODES.update(
    {
        "\n": "{ENTER}",
    }
)


class BaseProbe(FontMixin):
    fixed_height = None

    def __init__(self, native=None):
        self.native = native

    async def redraw(self, message=None, delay=0):
        """Request a redraw of the app, waiting until that redraw has completed."""
        # Winforms style changes always take effect immediately.

        # If we're running slow, wait for a second
        if toga.App.app.run_slow:
            delay = max(1, delay)
        if delay:
            print("Waiting for redraw" if message is None else message)

        # Sleep even if the delay is zero: this allows any pending callbacks on the
        # event loop to run.
        await asyncio.sleep(delay)

    @property
    def x(self):
        return round(self.native.Left / self.scale_factor)

    @property
    def y(self):
        return round(self.native.Top / self.scale_factor)

    @property
    def width(self):
        return round(self.native.Width / self.scale_factor)

    @property
    def height(self):
        return round(self.native.Height / self.scale_factor)

    def assert_width(self, min_width, max_width):
        assert min_width <= self.width <= max_width

    def assert_height(self, min_height, max_height):
        if self.fixed_height is not None:
            assert self.height == approx(self.fixed_height, rel=0.1)
        else:
            assert min_height <= self.height <= max_height

    @property
    def scale_factor(self):
        # For ScrollContainer
        if hasattr(self, "native_content"):
            return self.get_scale_factor(
                native_screen=Screen.FromControl(self.native_content)
            )
        # For Windows and others
        else:
            return self.get_scale_factor(native_screen=Screen.FromControl(self.native))

    def get_scale_factor(self, native_screen):
        screen_rect = wintypes.RECT(
            native_screen.Bounds.Left,
            native_screen.Bounds.Top,
            native_screen.Bounds.Right,
            native_screen.Bounds.Bottom,
        )
        windll.user32.MonitorFromRect.restype = c_void_p
        windll.user32.MonitorFromRect.argtypes = [wintypes.RECT, wintypes.DWORD]
        # MONITOR_DEFAULTTONEAREST = 2
        hMonitor = windll.user32.MonitorFromRect(screen_rect, 2)
        pScale = wintypes.UINT()
        windll.shcore.GetScaleFactorForMonitor(c_void_p(hMonitor), byref(pScale))
        return pScale.value / 100

    async def type_character(self, char, *, shift=False, ctrl=False, alt=False):
        try:
            key_code = KEY_CODES[char]
        except KeyError:
            assert len(char) == 1, char
            key_code = char

        if shift:
            key_code = "+" + key_code
        if ctrl:
            key_code = "^" + key_code
        if alt:
            key_code = "%" + key_code

        # This sends keys to the focused window, which isn't necessarily even in the
        # same app. Unfortunately that makes it difficult to run tests in the
        # background.
        SendKeys.SendWait(key_code)

    def assert_image_size(self, image_size, size, screen):
        scale_factor = self.get_scale_factor(native_screen=screen._impl.native)
        assert image_size == (
            round(size[0] * scale_factor),
            round(size[1] * scale_factor),
        )



from System.Drawing import Image as WinImage

from .probe import BaseProbe


class ImageProbe(BaseProbe):
    def __init__(self, app, image):
        super().__init__()
        self.app = app
        self.image = image
        assert isinstance(self.image._impl.native, WinImage)

    def supports_extension(self, extension):
        return extension.lower() in {".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff"}



import asyncio

from System import EventArgs
from System.Windows.Forms import (
    Form,
    FormBorderStyle,
    FormWindowState,
    MenuStrip,
    Panel,
    ToolStrip,
    ToolStripSeparator,
)

from toga import Size

from .dialogs import DialogsMixin
from .probe import BaseProbe


class WindowProbe(BaseProbe, DialogsMixin):
    # Disabling the close button requires overriding a protected method
    # (https://stackoverflow.com/a/7301828), which Python.NET doesn't support
    # (https://github.com/pythonnet/pythonnet/issues/2192).
    supports_closable = False
    supports_minimizable = True
    supports_move_while_hidden = True
    supports_unminimize = True
    supports_minimize = True
    supports_placement = True
    supports_as_image = True
    supports_focus = True

    def __init__(self, app, window):
        self.app = app
        self.window = window
        self.impl = window._impl
        super().__init__(window._impl.native)
        assert isinstance(self.native, Form)

    async def wait_for_window(
        self,
        message,
        state=None,
    ):
        await self.redraw(message)

        if state:
            timeout = 5
            polling_interval = 0.1
            exception = None
            loop = asyncio.get_running_loop()
            start_time = loop.time()
            while (loop.time() - start_time) < timeout:
                try:
                    assert self.instantaneous_state == state
                    return
                except AssertionError as e:
                    exception = e
                    await asyncio.sleep(polling_interval)
                    continue
                raise exception

    async def cleanup(self):
        self.window.close()
        await self.redraw("Closing window")

    def close(self):
        self.native.Close()

    @property
    def content_size(self):
        client_size = self.client_size
        return Size(
            client_size.width,
            client_size.height - (self.impl._top_bars_height() / self.scale_factor),
        )

    @property
    def client_size(self):
        return Size(
            self.native.ClientSize.Width / self.scale_factor,
            self.native.ClientSize.Height / self.scale_factor,
        )

    @property
    def is_resizable(self):
        return self.native.FormBorderStyle == FormBorderStyle.Sizable

    @property
    def is_minimizable(self):
        return self.native.MinimizeBox

    @property
    def is_minimized(self):
        return self.native.WindowState == FormWindowState.Minimized

    def minimize(self):
        if self.native.MinimizeBox:
            self.native.WindowState = FormWindowState.Minimized

    def unminimize(self):
        self.native.WindowState = FormWindowState.Normal

    @property
    def container_probe(self):
        panels = [
            control for control in self.native.Controls if isinstance(control, Panel)
        ]
        assert len(panels) == 1
        return BaseProbe(panels[0])

    @property
    def instantaneous_state(self):
        return self.impl.get_window_state(in_progress_state=False)

    @property
    def menubar_probe(self):
        return BaseProbe(bar) if (bar := self.native.MainMenuStrip) else None

    @property
    def toolbar_probe(self):
        return BaseProbe(bar) if (bar := self._native_toolbar()) else None

    def _native_toolbar(self):
        for control in self.native.Controls:
            if isinstance(control, ToolStrip) and not isinstance(control, MenuStrip):
                return control
        else:
            return None

    def has_toolbar(self):
        return self._native_toolbar() is not None

    def _native_toolbar_item(self, index):
        return self._native_toolbar().Items[index]

    def assert_is_toolbar_separator(self, index, section=False):
        assert isinstance(self._native_toolbar_item(index), ToolStripSeparator)

    def assert_toolbar_item(self, index, label, tooltip, has_icon, enabled):
        item = self._native_toolbar_item(index)
        assert item.Text == label
        assert item.ToolTipText == tooltip
        assert (item.Image is not None) == has_icon
        assert item.Enabled == enabled

    def press_toolbar_button(self, index):
        self._native_toolbar_item(index).OnClick(EventArgs.Empty)



from io import BytesIO

from PIL import Image
from System.Windows.Forms import MouseButtons, Panel

from .base import SimpleProbe


class CanvasProbe(SimpleProbe):
    native_class = Panel

    def reference_variant(self, reference):
        if reference in {"multiline_text", "write_text"}:
            return f"{reference}-winforms"
        return reference

    def get_image(self):
        return Image.open(BytesIO(self.impl.get_image_data()))

    async def mouse_press(self, x, y, **kwargs):
        self.native.OnMouseDown(self.mouse_event(x, y, **kwargs))
        self.native.OnMouseUp(self.mouse_event(x, y, **kwargs))

    async def mouse_activate(self, x, y, **kwargs):
        self.native.OnMouseDown(self.mouse_event(x, y, **kwargs))
        self.native.OnMouseUp(self.mouse_event(x, y, **kwargs))
        self.native.OnMouseDown(self.mouse_event(x, y, clicks=2, **kwargs))
        self.native.OnMouseUp(self.mouse_event(x, y, clicks=2, **kwargs))

    async def mouse_drag(self, x1, y1, x2, y2, **kwargs):
        # Without a mouse button pressed, a move event should be ignored.
        move_event = self.mouse_event((x1 + x2) // 2, (y1 + y2) // 2, **kwargs)
        self.native.OnMouseMove(move_event)

        self.native.OnMouseDown(self.mouse_event(x1, y1, **kwargs))
        self.native.OnMouseMove(move_event)
        self.native.OnMouseUp(self.mouse_event(x2, y2, **kwargs))

    async def alt_mouse_press(self, x, y):
        await self.mouse_press(x, y, button=MouseButtons.Right)

    async def alt_mouse_drag(self, x1, y1, x2, y2):
        await self.mouse_drag(x1, y1, x2, y2, button=MouseButtons.Right)



import pytest
from System.Windows.Forms import NumericUpDown

from .base import SimpleProbe
from .properties import toga_x_text_align


class NumberInputProbe(SimpleProbe):
    native_class = NumericUpDown
    allows_invalid_value = False
    allows_empty_value = True
    allows_extra_digits = True
    fixed_height = 23

    @property
    def value(self):
        return self.native.Text

    @property
    def readonly(self):
        return self.native.ReadOnly

    def clear_input(self):
        self.native.Text = ""

    async def increment(self):
        self.widget.focus()
        await self.type_character("<up>")

    async def decrement(self):
        self.widget.focus()
        await self.type_character("<down>")

    @property
    def text_align(self):
        return toga_x_text_align(self.native.TextAlign)

    def assert_vertical_text_align(self, expected):
        # Vertical text alignment isn't configurable in this native widget.
        pass

    def set_cursor_at_end(self):
        pytest.skip("Cursor positioning not supported on this platform")



from contextlib import contextmanager
from decimal import ROUND_UP

import System.Windows.Forms as WinForms

from ..libs.wrapper import WeakrefCallable
from .base import Widget


class TogaComboBox(WinForms.ComboBox):
    def __init__(self, impl):
        super().__init__()
        self.impl = impl
        self.DropDownStyle = WinForms.ComboBoxStyle.DropDownList
        self.SelectedIndexChanged += WeakrefCallable(
            self.winforms_selected_index_changed
        )

    def winforms_selected_index_changed(self, sender, event):
        self.impl.on_change()


class Selection(Widget):

    def create(self):
        self.native = TogaComboBox(self)
        self._send_notifications = True

    @contextmanager
    def suspend_notifications(self):
        self._send_notifications = False
        yield
        self._send_notifications = True

    def on_change(self):
        if self._send_notifications:
            self.interface.on_change()

    def clear(self):
        self.native.Items.Clear()
        self.on_change()

    def insert(self, index, item):
        self.native.Items.Insert(index, self.interface._title_for_item(item))

        # WinfForm.ComboBox does not select the first item, so it's done here.
        if self.native.SelectedIndex == -1:
            self.native.SelectedIndex = 0

    def change(self, item):
        index = self.interface._items.index(item)
        with self.suspend_notifications():
            self.insert(index, item)
            self.remove(index + 1, item)

        # Changing the item text can change the layout size.
        self.interface.refresh()

    def remove(self, index, item):
        selection_change = self.get_selected_index() == index
        self.native.Items.RemoveAt(index)

        # Removing the selected item will initially cause *nothing* to be selected.
        # Select an adjacent item if there is one.
        if selection_change:
            if self.native.Items.Count == 0:
                self.on_change()
            else:
                self.native.SelectedIndex = max(0, index - 1)

    def select_item(self, index, item):
        self.native.SelectedIndex = index

    def get_selected_index(self):
        index = self.native.SelectedIndex
        return None if index == -1 else index

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



from pathlib import Path

from System import (
    ArgumentException,
    OutOfMemoryException,
)
from System.Drawing import Image as WinImage
from System.Drawing.Imaging import ImageFormat
from System.IO import MemoryStream


class Image:
    RAW_TYPE = WinImage

    def __init__(self, interface, path=None, data=None, raw=None):
        self.interface = interface

        if path:
            try:
                self.native = WinImage.FromFile(str(path))
            except OutOfMemoryException:
                # OutOfMemoryException is what Winforms raises when a file
                # isn't a valid image file.
                raise ValueError(f"Unable to load image from {path}")
        elif data:
            try:
                stream = MemoryStream(data)
                self.native = WinImage.FromStream(stream)
            except ArgumentException:
                raise ValueError("Unable to load image from data")
        else:
            self.native = raw

    def get_width(self):
        return self.native.Width

    def get_height(self):
        return self.native.Height

    def get_data(self):
        stream = MemoryStream()
        self.native.Save(stream, ImageFormat.Png)
        return bytes(stream.ToArray())

    def save(self, path):
        path = Path(path)
        try:
            filetype = {
                ".jpg": ImageFormat.Jpeg,
                ".jpeg": ImageFormat.Jpeg,
                ".png": ImageFormat.Png,
                ".gif": ImageFormat.Gif,
                ".bmp": ImageFormat.Bmp,
                ".tiff": ImageFormat.Tiff,
            }[path.suffix.lower()]
            str_path = str(path)
        except KeyError:
            raise ValueError(f"Don't know how to save image of type {path.suffix!r}")

        self.native.Save(str_path, filetype)



import System.Windows.Forms
from pytest import xfail

from .base import SimpleProbe


class SelectionProbe(SimpleProbe):
    native_class = System.Windows.Forms.ComboBox
    fixed_height = 23

    def assert_resizes_on_content_change(self):
        xfail("Selection doesn't resize on content changes on this backend")

    @property
    def text_align(self):
        xfail("Can't change the text alignment of Selection on this backend")

    @property
    def titles(self):
        return list(self.native.Items)

    @property
    def selected_title(self):
        return self.native.SelectedItem

    async def select_item(self):
        self.native.SelectedIndex = 1



from System.Windows.Forms import TabControl, TabPage

from ..container import Container
from ..libs.wrapper import WeakrefCallable
from .base import Widget


class OptionContainer(Widget):
    uses_icons = False

    def create(self):
        self.native = TabControl()
        self.native.Selected += WeakrefCallable(self.winforms_selected)
        self.panels = []

    def add_option(self, index, text, widget, icon):
        page = TabPage(text)
        self.native.TabPages.Insert(index, page)

        panel = Container(page)
        self.panels.insert(index, panel)
        panel.set_content(widget)

        # ClientSize is set correctly for a newly-added tab, but is only updated on
        # resize for the selected tab. And when the selection changes, the
        # newly-selected tab's ClientSize is not updated until some time after the
        # Selected event fires.
        self.resize_content(panel)

        page.ClientSizeChanged += WeakrefCallable(self.winforms_client_size_changed)

    def remove_option(self, index):
        panel = self.panels.pop(index)
        panel.clear_content()

        self.native.TabPages.RemoveAt(index)

    def set_option_enabled(self, index, enabled):
        """Winforms documentation states that Enabled is not meaningful for this
        control.

        Disabling option only disables the content of the tab, not the tab itself.
        """
        self.native.TabPages[index].Enabled = enabled

    def is_option_enabled(self, index):
        return self.native.TabPages[index].Enabled

    def set_option_text(self, index, value):
        self.native.TabPages[index].Text = value

    def get_option_text(self, index):
        return self.native.TabPages[index].Text

    def set_option_icon(self, index, value):  # pragma: nocover
        # This shouldn't ever be invoked, but it's included for completeness.
        pass

    def get_option_icon(self, index):
        # Icons aren't supported
        return None

    def get_current_tab_index(self):
        return self.native.SelectedIndex

    def set_current_tab_index(self, current_tab_index):
        self.native.SelectedIndex = current_tab_index

    def winforms_selected(self, sender, event):
        self.interface.on_select()

    def winforms_client_size_changed(self, sender, event):
        for panel in self.panels:
            self.resize_content(panel)

    def resize_content(self, panel):
        size = panel.native_parent.ClientSize
        panel.resize_content(size.Width, size.Height)



import json
import webbrowser
from concurrent.futures import Future

import System.Windows.Forms as WinForms
from System import Action, String
from System.Drawing import Color
from System.Threading.Tasks import Task, TaskScheduler

import toga
from toga.types import LatLng
from toga_winforms.libs.extensions import (
    CoreWebView2CreationProperties,
    WebView2,
    WebView2RuntimeNotFoundException,
)

from ..libs.wrapper import WeakrefCallable
from .base import Widget

MAPVIEW_HTML_CONTENT = """<!DOCTYPE html>
<html lang="en">
<head>
    <link
        rel="stylesheet"
        href="https://unpkg.com/leaflet@1.9.4/dist/leaflet.css"
        integrity="sha256-p4NxAoJBhIIN+hmNHrzRCf9tD/miZyoHS5obTRR9BMY="
        crossorigin=""/>
    <script
        src="https://unpkg.com/leaflet@1.9.4/dist/leaflet.js"
        integrity="sha256-20nQCchB9co0qIjJZRGuk2/Z9VM+kNiyxNV1lvTlZBo="
        crossorigin=""></script>

    <style>
        html, body {
            height: 100%;
            margin: 0;
        }
        #map {
            height: 100%;
            width: 100%;
        }
    </style>
</head>
<body>
    <div id="map"></div>
    <script>
        const map = L.map("map");
        const pins = {};
        const tiles = L.tileLayer("https://tile.openstreetmap.org/{z}/{x}/{y}.png", {
            maxZoom: 20,
            attribution: '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>'
        }).addTo(map);
    </script>
</body>
</html>
"""  # noqa: E501


def pin_id(pin):
    "Rendering utility; output the ID of the pin"
    return hex(id(pin))


def latlng(location):
    "Rendering utility; output a lat/lng coordinate"
    return f"[{location.lat}, {location.lng}]"


def popup(pin):
    "Rendering utility; output the content of the pin popup"
    if pin.subtitle:
        return f"<b>{pin.title}</b><br>{pin.subtitle}"
    else:
        return f"<b>{pin.title}</b>"


class MapView(Widget):
    SUPPORTS_ON_SELECT = False

    def create(self):
        self.native = WebView2()
        # WebView has a 2 phase startup; the widget needs to be initialized,
        # the content needs to be loaded. We can't actually use the widget
        # until the content is fully loaded.
        self.native.CoreWebView2InitializationCompleted += WeakrefCallable(
            self.winforms_initialization_completed
        )
        self.native.NavigationCompleted += WeakrefCallable(
            self.winforms_navigation_completed
        )
        self.loaded_future = None

        props = CoreWebView2CreationProperties()
        props.UserDataFolder = str(toga.App.app.paths.cache / "WebView2")
        self.native.CreationProperties = props

        # Trigger the configuration of the webview
        self.corewebview2_available = None
        self.native.EnsureCoreWebView2Async(None)
        self.native.DefaultBackgroundColor = Color.Transparent

        self.backlog = []

    def winforms_initialization_completed(self, sender, args):
        # The WebView2 widget has an "internal" widget (CoreWebView2) that is
        # the actual web view. The view isn't ready until the internal widget has
        # completed initialization, and that isn't done until an explicit
        # request is made (EnsureCoreWebView2Async).
        if args.IsSuccess:
            # We've initialized, so we must have the runtime
            self.corewebview2_available = True
            settings = self.native.CoreWebView2.Settings
            self.default_user_agent = settings.UserAgent

            # To enable debugging,
            settings.AreDefaultContextMenusEnabled = False
            settings.AreDevToolsEnabled = False

            settings.IsZoomControlEnabled = True

            # Now that the widget is initialized, we can load the HTML content
            self.native.NavigateToString(MAPVIEW_HTML_CONTENT)

        elif isinstance(
            args.InitializationException, WebView2RuntimeNotFoundException
        ):  # pragma: nocover
            print("Could not find the Microsoft Edge WebView2 Runtime.")
            if self.corewebview2_available is None:
                # The initialize message is sent twice on failure.
                # We only want to show the dialog once, so track that we
                # know the runtime is missing.
                self.corewebview2_available = False
                WinForms.MessageBox.Show(
                    "The Microsoft Edge WebView2 Runtime is not installed. "
                    "Web content will not be displayed.\n\n"
                    "Click OK to download the WebView2 Evergreen Runtime "
                    "Bootstrapper from Microsoft.",
                    "Missing Edge Webview2 runtime",
                    WinForms.MessageBoxButtons.OK,
                    WinForms.MessageBoxIcon.Error,
                )
                webbrowser.open(
                    "https://developer.microsoft.com/en-us/microsoft-edge/webview2/#download"  # noqa: E501
                )

        else:  # pragma: nocover
            WinForms.MessageBox.Show(
                "A critical error has occurred and functionality may be impaired.\n\n"
                "The WebView2 initialization failed with an exception:\n\n"
                f"{args.InitializationException}",
                "Error",
                WinForms.MessageBoxButtons.OK,
                WinForms.MessageBoxIcon.Error,
            )

    def winforms_navigation_completed(self, sender, args):
        # The Toga API allows you to invoke methods on the MapView as soon
        # as it has been created; however, the web view implementing the map
        # can't process requests until the page has finished loading. As
        # soon as the page has finished loading, we can process the backlog
        # of API requests. Any return values will be ignored.
        for javascript in self.backlog:
            self.native.ExecuteScriptAsync(javascript)
        self.backlog = None

    def _invoke(self, javascript):
        if self.backlog is not None:
            self.backlog.append(javascript)
            return None
        else:
            future = Future()
            task_scheduler = TaskScheduler.FromCurrentSynchronizationContext()

            def callback(task):
                # If the evaluation fails, task.Result will be "null", with no way to
                # distinguish it from an actual null return value.
                value = json.loads(task.Result)
                future.set_result(value)

            self.native.ExecuteScriptAsync(javascript).ContinueWith(
                Action[Task[String]](callback), task_scheduler
            )

            # Process Winforms events until the result is available
            while not future.done():
                WinForms.Application.DoEvents()

            return future.result()

    def get_location(self):
        if self.backlog is None:
            result = self._invoke("map.getCenter();")
            return LatLng(result["lat"], result["lng"])
        else:  # pragma: no cover
            print(
                "MapView isn't fully initialized. "
                "Mapview.location result will be unreliable"
            )
            return LatLng(0.0, 0.0)

    def set_location(self, position):
        self._invoke(f"map.panTo({latlng(position)});")

    def get_zoom(self):
        return self._invoke("map.getZoom();")

    def set_zoom(self, zoom):
        self._invoke(f"map.setZoom({zoom});")

    def add_pin(self, pin):
        self._invoke(
            f'pins["{pin_id(pin)}"] = L.marker({latlng(pin.location)}).addTo(map)'
            f'.bindPopup("{popup(pin)}");'
        )

    def update_pin(self, pin):
        self._invoke(
            f'pins["{pin_id(pin)}"].setLatLng({latlng(pin.location)})'
            f'.setPopupContent("{popup(pin)}");'
        )

    def remove_pin(self, pin):
        self._invoke(
            f'map.removeLayer(pins["{pin_id(pin)}"]); delete pins["{pin_id(pin)}"];'
        )



from ctypes import c_uint, windll
from ctypes.wintypes import HWND, WPARAM
from decimal import ROUND_UP

import System.Windows.Forms as WinForms

from toga_winforms.colors import native_color
from toga_winforms.libs.fonts import HorizontalTextAlignment

from ..libs.wrapper import WeakrefCallable
from .base import Widget


class TextInput(Widget):

    def create(self):
        self.native = WinForms.TextBox()
        self.native.Multiline = False
        self.native.TextChanged += WeakrefCallable(self.winforms_text_changed)
        self.native.KeyPress += WeakrefCallable(self.winforms_key_press)
        self.native.GotFocus += WeakrefCallable(self.winforms_got_focus)
        self.native.LostFocus += WeakrefCallable(self.winforms_lost_focus)

        self._placeholder = ""

        self.error_provider = WinForms.ErrorProvider()
        self.error_provider.SetIconAlignment(
            self.native, WinForms.ErrorIconAlignment.MiddleRight
        )
        self.error_provider.SetIconPadding(self.native, -20)
        self.error_provider.BlinkStyle = WinForms.ErrorBlinkStyle.NeverBlink

    def get_readonly(self):
        return self.native.ReadOnly

    def set_readonly(self, value):
        self.native.ReadOnly = value

    def get_placeholder(self):
        return self._placeholder

    def set_placeholder(self, value):
        self._placeholder = value
        # This solution is based on https://stackoverflow.com/questions/4902565.
        EM_SETCUEBANNER = c_uint(0x1501)
        # value 0 means placeholder is hidden as soon the input gets focus
        # value 1 means placeholder is hidden only after something is typed into input
        show_placeholder_on_focus = WPARAM(1)
        window_handle = HWND(self.native.Handle.ToInt32())
        windll.user32.SendMessageW(
            window_handle,
            EM_SETCUEBANNER,
            show_placeholder_on_focus,
            self._placeholder,
        )

    def get_value(self):
        return self.native.Text

    def set_value(self, value):
        self.native.Text = value

    def set_text_align(self, value):
        self.native.TextAlign = HorizontalTextAlignment(value)

    def set_color(self, color):
        if color:
            self.native.ForeColor = native_color(color)
        else:
            self.native.ForeColor = self.native.DefaultForeColor

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )

    def winforms_text_changed(self, sender, event):
        self.interface._value_changed()

    def winforms_key_press(self, sender, event):
        if ord(event.KeyChar) == int(WinForms.Keys.Enter):
            self.interface.on_confirm()
            # Mark the event as handled; otherwise the Enter event will
            # propagate upstream, and generate a system beep.
            event.Handled = True

    def winforms_got_focus(self, sender, event):
        self.interface.on_gain_focus()

    def winforms_lost_focus(self, sender, event):
        self.interface.on_lose_focus()

    def is_valid(self):
        return self.error_provider.GetError(self.native) == ""

    def clear_error(self):
        self.error_provider.SetError(self.native, "")

    def set_error(self, error_message):
        self.error_provider.SetError(self.native, error_message)



import sys

from System import ArgumentException
from System.Drawing import Bitmap, Icon as WinIcon


class Icon:
    EXTENSIONS = [".ico", ".png", ".bmp"]
    SIZES = None

    def __init__(self, interface, path):
        self.interface = interface
        self.path = path

        try:
            if path is None:
                self.native = WinIcon.ExtractAssociatedIcon(sys.executable)
                self.bitmap = Bitmap.FromHicon(self.native.Handle)
            elif path.suffix == ".ico":
                self.native = WinIcon(str(path))
                self.bitmap = Bitmap.FromHicon(self.native.Handle)
            else:
                self.bitmap = Bitmap(str(path))
                self.native = WinIcon.FromHandle(self.bitmap.GetHicon())
        except ArgumentException:
            raise ValueError(f"Unable to load icon from {path}")



from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from System.Drawing import Bitmap

from toga.colors import TRANSPARENT
from toga.widgets.imageview import rehint_imageview

from .base import Widget


class ImageView(Widget):
    def create(self):
        self.native = WinForms.PictureBox()
        self._default_background_color = TRANSPARENT
        self.native.SizeMode = WinForms.PictureBoxSizeMode.Zoom

        # If self.native.Image is None, Winforms renders it as a white square
        # with a red cross through it. Ensure we always have an actual image,
        # using a 1x1 blank bitmap for the None case.
        self.native.Image = self._empty_image()

    def _empty_image(self):
        return Bitmap(1, 1)

    def set_image(self, image):
        # Destroy the existing image.
        self.native.Image.Dispose()

        if image:
            self.native.Image = self.interface._image._impl.native
        else:
            self.native.Image = self._empty_image()

    def rehint(self):
        width, height, aspect_ratio = rehint_imageview(
            self.interface.image, self.interface.style, self.dpi_scale
        )
        self.interface.intrinsic.width = self.scale_out(width, ROUND_UP)
        self.interface.intrinsic.height = self.scale_out(height, ROUND_UP)
        if aspect_ratio is not None:
            self.native.SizeMode = WinForms.PictureBoxSizeMode.Zoom
        else:
            self.native.SizeMode = WinForms.PictureBoxSizeMode.StretchImage



from System.Windows.Forms import TickStyle, TrackBar

from .base import SimpleProbe


class SliderProbe(SimpleProbe):
    native_class = TrackBar

    @property
    def position(self):
        return (self.native.Value - self._min) / (self._max - self._min)

    async def change(self, position):
        self.native.Value = self._min + round(position * (self._max - self._min))

    @property
    def tick_count(self):
        tick_style = self.native.TickStyle
        if tick_style == TickStyle.BottomRight:
            return self._max - self._min + 1
        elif tick_style == getattr(TickStyle, "None"):
            return None
        else:
            raise ValueError(f"unknown TickStyle {tick_style}")

    @property
    def _min(self):
        return self.native.Minimum

    @property
    def _max(self):
        return self.native.Maximum

    async def press(self):
        self.native.OnMouseDown(self.mouse_event())

    async def release(self):
        self.native.OnMouseUp(self.mouse_event())



from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from travertino.size import at_least

from toga.colors import TRANSPARENT

from ..libs.wrapper import WeakrefCallable
from .base import Widget


class Switch(Widget):
    def create(self):
        self.native = WinForms.CheckBox()
        self._default_background_color = TRANSPARENT
        self.native.CheckedChanged += WeakrefCallable(self.winforms_checked_changed)

    def winforms_checked_changed(self, sender, event):
        self.interface.on_change()

    def get_text(self):
        value = self.native.Text
        # Normalize a standalone ZERO WIDTH SPACE to an empty string.
        if value == "\u200b":
            return ""
        return value

    def set_text(self, text):
        if text == "":
            # An empty label would cause the widget's height to collapse, so display a
            # Unicode ZERO WIDTH SPACE instead.
            text = "\u200b"
        self.native.Text = text

    def get_value(self):
        return self.native.Checked

    def set_value(self, value):
        self.native.Checked = value

    def rehint(self):
        self.interface.intrinsic.width = self.scale_out(
            at_least(self.native.PreferredSize.Width), ROUND_UP
        )
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



import pytest
from System.Drawing import Bitmap
from System.Windows.Forms import (
    ColumnHeaderStyle,
    ListView,
    MouseButtons,
    MouseEventArgs,
)

from .base import SimpleProbe


class TableProbe(SimpleProbe):
    native_class = ListView
    supports_icons = 1  # First column only
    supports_keyboard_shortcuts = False
    supports_keyboard_boundary_shortcuts = True
    supports_widgets = False

    @property
    def row_count(self):
        return self.native.VirtualListSize

    @property
    def column_count(self):
        return len(self.native.Columns)

    def assert_cell_content(self, row, col, value=None, icon=None, widget=None):
        if widget:
            pytest.skip("This backend doesn't support widgets in Tables")
        else:
            lvi = self.native.Items[row]
            assert lvi.SubItems[col].Text == value
            if col == 0:
                if icon is None:
                    assert lvi.ImageIndex == -1
                    assert lvi.ImageKey == ""
                else:
                    imagelist = self.native.SmallImageList
                    size = imagelist.ImageSize
                    assert size.Width == size.Height == 16

                    # The image is resized and copied, so we need to compare the actual
                    # pixels.
                    actual = imagelist.Images[lvi.ImageIndex]
                    expected = Bitmap(icon._impl.bitmap, size)
                    for x in range(size.Width):
                        for y in range(size.Height):
                            assert actual.GetPixel(x, y) == expected.GetPixel(x, y)
            else:
                assert icon is None

    @property
    def max_scroll_position(self):
        document_height = (
            self.native.Items[self.row_count - 1].Bounds.Bottom
            - self.native.Items[0].Bounds.Top
        )
        return round(
            (document_height - self.native.ClientSize.Height) / self.scale_factor
        )

    @property
    def scroll_position(self):
        return -round((self.native.Items[0].Bounds.Top) / self.scale_factor)

    async def wait_for_scroll_completion(self):
        # No animation associated with scroll, so this is a no-op
        pass

    @property
    def header_visible(self):
        return self.native.HeaderStyle != getattr(ColumnHeaderStyle, "None")

    @property
    def header_titles(self):
        return [col.Text for col in self.native.Columns]

    def column_width(self, index):
        return round(self.native.Columns[index].Width / self.scale_factor)

    async def select_row(self, row, add=False):
        item = self.native.Items[row]
        if add:
            item.Selected = not item.Selected
        else:
            item.Selected = True

    async def activate_row(self, row):
        await self.select_row(row)

        bounds = self.native.Items[row].Bounds
        self.native.OnMouseDoubleClick(
            MouseEventArgs(
                MouseButtons.Left,
                clicks=2,
                x=int((bounds.Left + bounds.Right) / 2),
                y=int((bounds.Top + bounds.Bottom) / 2),
                delta=0,
            )
        )

    async def acquire_keyboard_focus(self):
        await self.type_character(
            "\t"
        )  # switch to widget.focus() when possible (#2972)
        await self.type_character(" ")  # select first row



from System.Drawing import FontFamily, SystemFonts

from toga.fonts import (
    BOLD,
    CURSIVE,
    FANTASY,
    ITALIC,
    MESSAGE,
    MONOSPACE,
    NORMAL,
    OBLIQUE,
    SANS_SERIF,
    SERIF,
    SMALL_CAPS,
    SYSTEM,
    SYSTEM_DEFAULT_FONT_SIZE,
)


class FontMixin:
    supports_custom_fonts = True
    supports_custom_variable_fonts = True

    @property
    def font(self):
        return self.native.Font

    def assert_font_options(self, weight=NORMAL, style=NORMAL, variant=NORMAL):
        assert BOLD if self.font.Bold else NORMAL == weight

        if style == OBLIQUE:
            print("Interpreting OBLIQUE font as ITALIC")
            assert self.font.Italic
        else:
            assert (ITALIC if self.font.Italic else NORMAL) == style

        if variant == SMALL_CAPS:
            print("Ignoring SMALL CAPS font test")
        else:
            assert NORMAL == variant

    @property
    def font_size(self):
        return round(self.font.SizeInPoints / self.scale_factor)

    def assert_font_size(self, expected):
        if expected == SYSTEM_DEFAULT_FONT_SIZE:
            expected = 9
        assert self.font_size == expected

    def assert_font_family(self, expected):
        assert str(self.font.Name) == {
            CURSIVE: "Comic Sans MS",
            FANTASY: "Impact",
            MESSAGE: SystemFonts.MessageBoxFont.FontFamily.Name,
            MONOSPACE: FontFamily.GenericMonospace.Name,
            SANS_SERIF: FontFamily.GenericSansSerif.Name,
            SERIF: FontFamily.GenericSerif.Name,
            SYSTEM: SystemFonts.MessageBoxFont.FontFamily.Name,
        }.get(expected, expected)



import asyncio
from pathlib import Path

import System.Windows.Forms as WinForms
from System.Drawing import (
    ContentAlignment,
    Font as WinFont,
    FontFamily,
)
from System.Windows.Forms import DialogResult, MessageBoxButtons, MessageBoxIcon

from .libs.user32 import DPI_AWARENESS_CONTEXT_UNAWARE, SetThreadDpiAwarenessContext
from .libs.wrapper import WeakrefCallable


class BaseDialog:
    def show(self, host_window, future):
        self.future = future

        # Don't differentiate between app and window modal dialogs
        # Show the dialog using an inner loop.
        asyncio.get_event_loop().start_inner_loop(self._show)


class MessageDialog(BaseDialog):
    def __init__(
        self,
        title,
        message,
        buttons,
        icon,
        success_result=None,
    ):
        super().__init__()
        self.message = message
        self.title = title
        self.buttons = buttons
        self.icon = icon
        self.success_result = success_result

    def _show(self):
        return_value = WinForms.MessageBox.Show(
            self.message,
            self.title,
            self.buttons,
            self.icon,
        )
        if self.success_result:
            self.future.set_result(return_value == self.success_result)
        else:
            self.future.set_result(None)


class InfoDialog(MessageDialog):
    def __init__(self, title, message):
        super().__init__(
            title,
            message,
            MessageBoxButtons.OK,
            MessageBoxIcon.Information,
        )


class QuestionDialog(MessageDialog):
    def __init__(self, title, message):
        super().__init__(
            title,
            message,
            MessageBoxButtons.YesNo,
            MessageBoxIcon.Information,
            success_result=DialogResult.Yes,
        )


class ConfirmDialog(MessageDialog):
    def __init__(self, title, message):
        super().__init__(
            title,
            message,
            MessageBoxButtons.OKCancel,
            MessageBoxIcon.Warning,
            success_result=DialogResult.OK,
        )


class ErrorDialog(MessageDialog):
    def __init__(self, title, message=None):
        super().__init__(
            title,
            message,
            WinForms.MessageBoxButtons.OK,
            WinForms.MessageBoxIcon.Error,
        )


class StackTraceDialog(BaseDialog):
    def __init__(self, title, message, content, retry):
        super().__init__()

        # This dialog uses a fixed layout, so we create it as DPI-unaware so it will be
        # scaled by the system. "When a window is created, its DPI awareness is defined
        # as the DPI awareness of the calling thread at that time."
        # (https://learn.microsoft.com/en-us/windows/win32/hidpi/high-dpi-improvements-for-desktop-applications).
        self.prev_dpi_context = None
        if SetThreadDpiAwarenessContext is not None:  # pragma: no branch
            self.prev_dpi_context = SetThreadDpiAwarenessContext(
                DPI_AWARENESS_CONTEXT_UNAWARE
            )
            if not self.prev_dpi_context:  # pragma: no cover
                print("WARNING: Failed to set DPI Awareness for StackTraceDialog")

        # Changing the DPI awareness causes confusion around font sizes, so set them
        # all explicitly.
        font_size = 8.25
        message_font = WinFont(FontFamily.GenericSansSerif, font_size)
        monospace_font = WinFont(FontFamily.GenericMonospace, font_size)

        self.native = WinForms.Form()
        self.native.MinimizeBox = False
        self.native.FormBorderStyle = self.native.FormBorderStyle.FixedSingle
        self.native.MaximizeBox = False
        self.native.FormClosing += WeakrefCallable(self.winforms_FormClosing)
        self.native.Width = 540
        self.native.Height = 320
        self.native.Text = title
        self.title = title

        # The top-of-page introductory message
        textLabel = WinForms.Label()
        textLabel.Left = 10
        textLabel.Top = 10
        textLabel.Width = 520
        textLabel.Alignment = ContentAlignment.MiddleCenter
        textLabel.Text = message
        textLabel.Font = message_font
        self.native.Controls.Add(textLabel)

        # A scrolling text box for the stack trace.
        trace = WinForms.RichTextBox()
        trace.Left = 10
        trace.Top = 35
        trace.Width = 504
        trace.Height = 205
        trace.Multiline = True
        trace.ReadOnly = True
        trace.Font = monospace_font
        trace.Text = content

        self.native.Controls.Add(trace)

        # Add acceptance/close buttons
        if retry:
            retry = WinForms.Button()
            retry.Left = 290
            retry.Top = 250
            retry.Width = 100
            retry.Text = "&Retry"
            retry.Font = message_font
            retry.Click += WeakrefCallable(self.winforms_Click_retry)

            self.native.Controls.Add(retry)

            quit = WinForms.Button()
            quit.Left = 400
            quit.Top = 250
            quit.Width = 100
            quit.Text = "&Quit"
            quit.Font = message_font
            quit.Click += WeakrefCallable(self.winforms_Click_quit)

            self.native.Controls.Add(quit)
        else:
            accept = WinForms.Button()
            accept.Left = 400
            accept.Top = 250
            accept.Width = 100
            accept.Text = "&OK"
            accept.Font = message_font
            accept.Click += WeakrefCallable(self.winforms_Click_accept)

            self.native.Controls.Add(accept)

    def _show(self):
        self.native.ShowDialog()

    def winforms_FormClosing(self, sender, event):
        # If the close button is pressed, the future won't be done.
        # We cancel this event to prevent the dialog from closing.
        # If a button is pressed, the future will be set, and a close
        # event will be triggered.
        if not self.future.done():
            event.Cancel = True  # pragma: no cover
        else:
            # Reverting the DPI awareness at the end of __init__ would cause the window
            # to be DPI-aware, presumably because the window isn't actually "created"
            # until we call ShowDialog.
            #
            # This cleanup doesn't make any difference to the dialogs example, because
            # "When the window procedure for a window is called [e.g. when clicking a
            # button], the thread is automatically switched to the DPI awareness context
            # that was in use when the window was created." However, other apps may do
            # things outside of the context of a window event.
            if self.prev_dpi_context:  # pragma: no branch
                SetThreadDpiAwarenessContext(self.prev_dpi_context)

    def winforms_Click_quit(self, sender, event):
        self.future.set_result(False)
        self.native.Close()

    def winforms_Click_retry(self, sender, event):
        self.future.set_result(True)
        self.native.Close()

    def winforms_Click_accept(self, sender, event):
        self.future.set_result(None)
        self.native.Close()


class FileDialog(BaseDialog):
    def __init__(
        self,
        native,
        title,
        initial_directory,
        *,
        filename=None,
        file_types=None,
    ):
        super().__init__()
        self.title = title
        self.native = native

        self._set_title(title)
        if filename is not None:
            native.FileName = filename

        if initial_directory is not None:
            self._set_initial_directory(str(initial_directory))

        if file_types is not None:
            filters = [f"{ext} files (*.{ext})|*.{ext}" for ext in file_types] + [
                "All files (*.*)|*.*"
            ]

            if len(file_types) > 1:
                pattern = ";".join([f"*.{ext}" for ext in file_types])
                filters.insert(0, f"All matching files ({pattern})|{pattern}")

            native.Filter = "|".join(filters)

    def _show(self):
        response = self.native.ShowDialog()
        if response == DialogResult.OK:
            self.future.set_result(self._get_filenames())
        else:
            self.future.set_result(None)

    def _set_title(self, title):
        self.native.Title = title

    def _set_initial_directory(self, initial_directory):
        self.native.InitialDirectory = initial_directory


class SaveFileDialog(FileDialog):
    def __init__(self, title, filename, initial_directory, file_types):
        super().__init__(
            WinForms.SaveFileDialog(),
            title,
            initial_directory,
            filename=filename,
            file_types=file_types,
        )

    def _get_filenames(self):
        return Path(self.native.FileName)


class OpenFileDialog(FileDialog):
    def __init__(
        self,
        title,
        initial_directory,
        file_types,
        multiple_select,
    ):
        super().__init__(
            WinForms.OpenFileDialog(),
            title,
            initial_directory,
            file_types=file_types,
        )
        if multiple_select:
            self.native.Multiselect = True

    # Provided as a stub that can be mocked in test conditions
    def selected_paths(self):
        return self.native.FileNames

    def _get_filenames(self):
        if self.native.Multiselect:
            return [Path(filename) for filename in self.selected_paths()]
        else:
            return Path(self.native.FileName)


class SelectFolderDialog(FileDialog):
    def __init__(self, title, initial_directory, multiple_select):
        super().__init__(
            WinForms.FolderBrowserDialog(),
            title,
            initial_directory,
        )

        # The native dialog doesn't support multiple selection, so the only effect
        # this has is to change whether we return a list.
        self.multiple_select = multiple_select

    def _get_filenames(self):
        filename = Path(self.native.SelectedPath)
        return [filename] if self.multiple_select else filename

    def _set_title(self, title):
        self.native.Description = title

    def _set_initial_directory(self, initial_directory):
        self.native.SelectedPath = initial_directory



import System.Windows.Forms

from .base import SimpleProbe
from .properties import toga_x_text_align, toga_y_text_align


class LabelProbe(SimpleProbe):
    native_class = System.Windows.Forms.Label

    @property
    def text(self):
        return self.native.Text

    @property
    def text_align(self):
        return toga_x_text_align(self.native.TextAlign)

    def assert_vertical_text_align(self, expected):
        assert toga_y_text_align(self.native.TextAlign) == expected



import sys
from decimal import ROUND_UP, Decimal, InvalidOperation

import System.Windows.Forms as WinForms
from System import Convert, String

from toga.widgets.numberinput import _clean_decimal
from toga_winforms.libs.fonts import HorizontalTextAlignment

from ..libs.wrapper import WeakrefCallable
from .base import Widget


def native_decimal(value):
    if isinstance(value, Decimal):
        # The explicit type is needed to prevent single-character strings from calling
        # the Char overload, which always throws an exception.
        return Convert.ToDecimal.__overloads__[String](str(value))
    else:
        assert isinstance(value, int)
        return Convert.ToDecimal(value)


class NumberInput(Widget):

    def create(self):
        self.native = WinForms.NumericUpDown()
        self.native.TextChanged += WeakrefCallable(self.winforms_text_changed)

    def winforms_text_changed(self, sender, event):
        self.interface.on_change()

    def get_readonly(self):
        return self.native.ReadOnly

    def set_readonly(self, value):
        self.native.ReadOnly = value

    def set_step(self, step):
        self.native.Increment = native_decimal(step)
        self.native.DecimalPlaces = abs(step.as_tuple().exponent)

    def set_min_value(self, value):
        self.native.Minimum = native_decimal(
            (-sys.maxsize - 1) if value is None else value,
        )

    def set_max_value(self, value):
        self.native.Maximum = native_decimal(
            sys.maxsize if value is None else value,
        )

    def get_value(self):
        try:
            return _clean_decimal(self.native.Text, self.interface.step)
        except InvalidOperation:
            return None

    def set_value(self, value):
        if value is None:
            self.native.Text = ""
        else:
            self.native.Value = native_decimal(value)

    def set_text_align(self, value):
        self.native.TextAlign = HorizontalTextAlignment(value)

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



import json
import webbrowser
from http.cookiejar import Cookie, CookieJar

import System.Windows.Forms as WinForms
from System import (
    Action,
    String,
    Uri,
)
from System.Collections.Generic import List  # Import List for generics
from System.Drawing import Color
from System.Threading.Tasks import Task, TaskScheduler

import toga
from toga.widgets.webview import CookiesResult, JavaScriptResult
from toga_winforms.libs.extensions import (
    CoreWebView2Cookie,
    CoreWebView2CreationProperties,
    WebView2,
    WebView2RuntimeNotFoundException,
)

from ..libs.wrapper import WeakrefCallable
from .base import Widget


def requires_initialization(method):
    def wrapper(self, *args, **kwargs):
        def task():
            method(self, *args, **kwargs)

        self.run_after_initialization(task)

    return wrapper


def cookies_completion_handler(result):
    """
    Generalized completion handler for processing cookies.
    """

    def _completion_handler(task):

        # Initialize a CookieJar to store cookies
        cookie_jar = CookieJar()

        # Use a list comprehension to convert each cookie into a Cookie
        # object and add it to the CookieJar
        [
            cookie_jar.set_cookie(
                Cookie(
                    version=0,
                    name=cookie.Name,
                    value=cookie.Value,
                    port=None,
                    port_specified=False,
                    domain=cookie.Domain,
                    domain_specified=True,
                    domain_initial_dot=False,
                    path=cookie.Path,
                    path_specified=True,
                    secure=cookie.IsSecure,
                    expires=None,
                    discard=cookie.IsSession,
                    comment=None,
                    comment_url=None,
                    rest={},
                    rfc2109=False,  # Whether the cookie follows RFC 2109
                )
            )
            for cookie in task.Result
        ]

        # Set the CookieJar in the CookiesResult object
        result.set_result(cookie_jar)

    return _completion_handler


class WebView(Widget):
    def create(self):
        self.native = WebView2()
        self.native.CoreWebView2InitializationCompleted += WeakrefCallable(
            self.winforms_initialization_completed
        )
        self.native.NavigationCompleted += WeakrefCallable(
            self.winforms_navigation_completed
        )
        self.loaded_future = None

        props = CoreWebView2CreationProperties()
        props.UserDataFolder = str(toga.App.app.paths.cache / "WebView2")
        self.native.CreationProperties = props

        # Trigger the configuration of the webview
        self.corewebview2_available = None
        self.pending_tasks = []
        self.native.EnsureCoreWebView2Async(None)
        self.native.DefaultBackgroundColor = Color.Transparent

    # Any non-trivial use of the WebView requires the CoreWebView2 object to be
    # initialized, which is asynchronous. Since most of this class's methods are not
    # asynchronous, they cannot handle this using `await`. Instead, they add a callable
    # to a queue of pending tasks, which is processed once we receive the
    # CoreWebView2InitializationCompleted event.
    def run_after_initialization(self, task):
        if self.corewebview2_available:
            task()
        else:
            self.pending_tasks.append(task)

    def winforms_initialization_completed(self, sender, args):
        # The WebView2 widget has an "internal" widget (CoreWebView2) that is
        # the actual web view. The view isn't ready until the internal widget has
        # completed initialization, and that isn't done until an explicit
        # request is made (EnsureCoreWebView2Async).
        if args.IsSuccess:
            # We've initialized, so we must have the runtime
            self.corewebview2_available = True
            settings = self.native.CoreWebView2.Settings
            self.default_user_agent = settings.UserAgent

            # Initialize cookie manager
            self.cookie_manager = self.native.CoreWebView2.CookieManager

            debug = True
            settings.AreBrowserAcceleratorKeysEnabled = debug
            settings.AreDefaultContextMenusEnabled = debug
            settings.AreDefaultScriptDialogsEnabled = True
            settings.AreDevToolsEnabled = debug
            settings.IsBuiltInErrorPageEnabled = True
            settings.IsScriptEnabled = True
            settings.IsWebMessageEnabled = True
            settings.IsStatusBarEnabled = debug
            settings.IsSwipeNavigationEnabled = False
            settings.IsZoomControlEnabled = True

            for task in self.pending_tasks:
                task()
            self.pending_tasks = None

        elif isinstance(
            args.InitializationException, WebView2RuntimeNotFoundException
        ):  # pragma: nocover
            print("Could not find the Microsoft Edge WebView2 Runtime.")
            if self.corewebview2_available is None:
                # The initialize message is sent twice on failure.
                # We only want to show the dialog once, so track that we
                # know the runtime is missing.
                self.corewebview2_available = False
                WinForms.MessageBox.Show(
                    "The Microsoft Edge WebView2 Runtime is not installed. "
                    "Web content will not be displayed.\n\n"
                    "Click OK to download the WebView2 Evergreen Runtime "
                    "Bootstrapper from Microsoft.",
                    "Missing Edge Webview2 runtime",
                    WinForms.MessageBoxButtons.OK,
                    WinForms.MessageBoxIcon.Error,
                )
                webbrowser.open(
                    "https://developer.microsoft.com/en-us/microsoft-edge/webview2/#download"  # noqa: E501
                )

        else:  # pragma: nocover
            WinForms.MessageBox.Show(
                "A critical error has occurred and functionality may be impaired.\n\n"
                "The WebView2 initialization failed with an exception:\n\n"
                f"{args.InitializationException}",
                "Error",
                WinForms.MessageBoxButtons.OK,
                WinForms.MessageBoxIcon.Error,
            )

    def winforms_navigation_completed(self, sender, args):
        self.interface.on_webview_load()

        if self.loaded_future:
            self.loaded_future.set_result(None)
            self.loaded_future = None

    def get_url(self):
        source = self.native.Source
        if source is None:  # pragma: nocover
            return None  # CoreWebView2 is not yet initialized.
        else:
            url = str(source)
            return None if url == "about:blank" else url

    @requires_initialization
    def set_url(self, value, future=None):
        self.loaded_future = future
        if value is None:
            self.set_content("about:blank", "")
        else:
            self.native.Source = Uri(value)

    @requires_initialization
    def set_content(self, root_url, content):
        # There appears to be no way to pass the root_url.
        self.native.NavigateToString(content)

    def get_user_agent(self):
        if self.corewebview2_available:
            return self.native.CoreWebView2.Settings.UserAgent
        else:  # pragma: nocover
            return ""

    @requires_initialization
    def set_user_agent(self, value):
        self.native.CoreWebView2.Settings.UserAgent = (
            self.default_user_agent if value is None else value
        )

    def get_cookies(self):
        """
        Retrieve all cookies asynchronously from the WebView.

        :return: A CookiesResult object that can be awaited.
        """
        # Create an AsyncResult to manage the cookies
        result = CookiesResult()

        # Wrap the Python completion handler in a .NET Action delegate
        completion_handler_delegate = Action[Task[List[CoreWebView2Cookie]]](
            cookies_completion_handler(result)
        )

        # Call the method to retrieve cookies asynchronously
        task_scheduler = TaskScheduler.FromCurrentSynchronizationContext()
        self.cookie_manager.GetCookiesAsync(None).ContinueWith(
            completion_handler_delegate, task_scheduler
        )

        return result

    def evaluate_javascript(self, javascript, on_result=None):
        result = JavaScriptResult(on_result)
        task_scheduler = TaskScheduler.FromCurrentSynchronizationContext()

        def callback(task):
            # If the evaluation fails, task.Result will be "null", with no way to
            # distinguish it from an actual null return value.
            value = json.loads(task.Result)
            result.set_result(value)

        def execute():
            self.native.ExecuteScriptAsync(javascript).ContinueWith(
                Action[Task[String]](callback), task_scheduler
            )

        self.run_after_initialization(execute)
        return result



from warnings import warn

import System.Windows.Forms as WinForms

import toga

from ..libs.wrapper import WeakrefCallable
from .base import Widget


class Table(Widget):

    # The following methods are overridden in DetailedList.
    @property
    def _headings(self):
        return self.interface.headings

    @property
    def _accessors(self):
        return self.interface.accessors

    @property
    def _multiple_select(self):
        return self.interface.multiple_select

    @property
    def _data(self):
        return self.interface.data

    def create(self):
        self.native = WinForms.ListView()
        self.native.View = WinForms.View.Details
        self._cache = []
        self._first_item = 0
        self._pending_resize = True

        headings = self._headings
        self.native.HeaderStyle = (
            getattr(WinForms.ColumnHeaderStyle, "None")
            if headings is None
            else WinForms.ColumnHeaderStyle.Nonclickable
        )

        dataColumn = []
        for i, accessor in enumerate(self._accessors):
            heading = None if headings is None else headings[i]
            dataColumn.append(self._create_column(heading, accessor))

        self.native.FullRowSelect = True
        self.native.MultiSelect = self._multiple_select
        self.native.DoubleBuffered = True
        self.native.VirtualMode = True
        self.native.Columns.AddRange(dataColumn)
        self.native.SmallImageList = WinForms.ImageList()
        self.native.HideSelection = False

        self.native.ItemSelectionChanged += WeakrefCallable(
            self.winforms_item_selection_changed
        )
        self.native.RetrieveVirtualItem += WeakrefCallable(
            self.winforms_retrieve_virtual_item
        )
        self.native.CacheVirtualItems += WeakrefCallable(
            self.winforms_cache_virtual_items
        )
        self.native.SearchForVirtualItem += WeakrefCallable(
            self.winforms_search_for_virtual_item
        )
        self.native.VirtualItemsSelectionRangeChanged += WeakrefCallable(
            self.winforms_item_selection_changed
        )
        self.add_action_events()

    def add_action_events(self):
        self.native.MouseDoubleClick += WeakrefCallable(self.winforms_double_click)

    def set_bounds(self, x, y, width, height):
        super().set_bounds(x, y, width, height)
        if self._pending_resize:
            self._pending_resize = False
            self._resize_columns()

    def winforms_retrieve_virtual_item(self, sender, e):
        # Because ListView is in VirtualMode, it's necessary implement
        # VirtualItemsSelectionRangeChanged event to create ListViewItem
        # when it's needed
        if (
            self._cache
            and e.ItemIndex >= self._first_item
            and e.ItemIndex < self._first_item + len(self._cache)
        ):
            e.Item = self._cache[e.ItemIndex - self._first_item]
        else:
            e.Item = self._new_item(e.ItemIndex)

    def winforms_cache_virtual_items(self, sender, e):
        if (
            self._cache
            and e.StartIndex >= self._first_item
            and e.EndIndex < self._first_item + len(self._cache)
        ):
            # If the newly requested cache is a subset of the old cache,
            # no need to rebuild everything, so do nothing
            return

        # Now we need to rebuild the cache.
        self._first_item = e.StartIndex
        new_length = e.EndIndex - e.StartIndex + 1
        self._cache = []

        # Fill the cache with the appropriate ListViewItems.
        for i in range(new_length):
            self._cache.append(self._new_item(i + self._first_item))

    def winforms_search_for_virtual_item(self, sender, e):
        if (
            not e.IsTextSearch or not self._accessors or not self._data
        ):  # pragma: no cover
            # If this list is empty, or has no columns, or it's an unsupported search
            # type, there's no search to be done. These situation are difficult to
            # trigger in CI; they're here as a safety catch.
            return
        find_previous = e.Direction in [
            WinForms.SearchDirectionHint.Up,
            WinForms.SearchDirectionHint.Left,
        ]
        i = e.StartIndex
        found_item = False
        while True:
            # It is possible for e.StartIndex to be received out-of-range if the user
            # performs keyboard navigation at its edge, so check before accessing data
            if i < 0:  # pragma: no cover
                # This could happen if this event is fired searching backwards,
                # however this should not happen in Toga's use of it.
                # i = len(self._data) - 1
                raise NotImplementedError("backwards search unsupported")
            elif i >= len(self._data):
                i = 0
            if (
                self._item_text(self._data[i], self._accessors[0])[
                    : len(e.Text)
                ].lower()
                == e.Text.lower()
            ):
                found_item = True
                break
            if find_previous:  # pragma: no cover
                # Toga does not currently need backwards searching functionality.
                # i -= 1
                raise NotImplementedError("backwards search unsupported")
            else:
                i += 1
            if i == e.StartIndex:
                break
        if found_item:
            e.Index = i

    def winforms_item_selection_changed(self, sender, e):
        self.interface.on_select()

    def winforms_double_click(self, sender, e):
        hit_test = self.native.HitTest(e.X, e.Y)
        item = hit_test.Item
        if item is not None:
            self.interface.on_activate(row=self._data[item.Index])
        else:  # pragma: no cover
            # Double clicking outside of an item apparently doesn't raise the event, but
            # that isn't guaranteed by the documentation.
            pass

    def _create_column(self, heading, accessor):
        col = WinForms.ColumnHeader()
        col.Text = heading
        col.Name = accessor
        return col

    def _resize_columns(self):
        num_cols = len(self.native.Columns)
        if num_cols == 0:
            return

        width = int(self.native.ClientSize.Width / num_cols)
        for col in self.native.Columns:
            col.Width = width

    def change_source(self, source):
        self.update_data()

    def _new_item(self, index):
        item = self._data[index]

        def icon(attr):
            val = getattr(item, attr, None)
            icon = None
            if isinstance(val, tuple):
                if val[0] is not None:
                    icon = val[0]
            else:
                try:
                    icon = val.icon
                except AttributeError:
                    pass

            return None if icon is None else icon._impl

        lvi = WinForms.ListViewItem(
            [self._item_text(item, attr) for attr in self._accessors],
        )

        # If the table has accessors, populate the icons for the table.
        if self._accessors:
            # TODO: ListView only has built-in support for one icon per row. One
            # possible workaround is in https://stackoverflow.com/a/46128593.
            icon = icon(self._accessors[0])
            if icon is not None:
                lvi.ImageIndex = self._image_index(icon)

        return lvi

    def _item_text(self, item, attr):
        val = getattr(item, attr, None)
        if isinstance(val, toga.Widget):
            warn("Winforms does not support the use of widgets in cells")
            val = None
        if isinstance(val, tuple):
            val = val[1]
        if val is None:
            val = self.interface.missing_value
        return str(val)

    def _image_index(self, icon):
        images = self.native.SmallImageList.Images
        key = str(icon.path)
        index = images.IndexOfKey(key)
        if index == -1:
            index = images.Count
            images.Add(key, icon.bitmap)
        return index

    def update_data(self):
        self.native.VirtualListSize = len(self._data)
        self._cache = []

    def insert(self, index, item):
        self.update_data()

    def change(self, item):
        self.update_data()

    def remove(self, index, item):
        self.update_data()

    def clear(self):
        self.update_data()

    def get_selection(self):
        selected_indices = list(self.native.SelectedIndices)
        if self._multiple_select:
            return selected_indices
        elif len(selected_indices) == 0:
            return None
        else:
            return selected_indices[0]

    def scroll_to_row(self, index):
        self.native.EnsureVisible(index)

    def remove_column(self, index):
        self.native.Columns.RemoveAt(index)
        self.update_data()
        self._resize_columns()

    def insert_column(self, index, heading, accessor):
        self.native.Columns.Insert(index, self._create_column(heading, accessor))
        self.update_data()
        self._resize_columns()



from abc import ABC, abstractmethod
from decimal import ROUND_HALF_EVEN, Decimal

from System.Drawing import (
    Point,
    Size,
    SystemColors,
)
from travertino.size import at_least

from toga.colors import TRANSPARENT, rgba
from toga_winforms.colors import (
    native_color,
    toga_color,
)


class Scalable(ABC):
    SCALE_DEFAULT_ROUNDING = ROUND_HALF_EVEN

    @property
    @abstractmethod
    def dpi_scale(self):
        raise NotImplementedError()

    # Convert CSS pixels to native pixels
    def scale_in(self, value, rounding=SCALE_DEFAULT_ROUNDING):
        return self.scale_round(value * self.dpi_scale, rounding)

    # Convert native pixels to CSS pixels
    def scale_out(self, value, rounding=SCALE_DEFAULT_ROUNDING):
        if isinstance(value, at_least):
            return at_least(self.scale_out(value.value, rounding))
        else:
            return self.scale_round(value / self.dpi_scale, rounding)

    def scale_round(self, value, rounding):
        if rounding is None:
            return value
        return int(Decimal(value).to_integral(rounding))


class Widget(Scalable, ABC):
    def __init__(self, interface):
        self.interface = interface

        self._container = None
        self.native = None
        self.create()

        # Widgets that need to set a different default background_color should override
        # the _default_background_color attribute.
        #
        # Note: On Winforms, _default_background_color is set in the form of toga color,
        #       instead of the native Color. This is because we need to manually do the
        #       alpha blending, and the native Color class does not directly handle the
        #       alpha transparency in the same way.
        if not hasattr(self, "_default_background_color"):
            # If a widget hasn't specifically defined a default background color then
            # set the system assigned background color as the default background color
            # of the widget.
            self._default_background_color = toga_color(self.native.BackColor)

        # Obtain a Graphics object and immediately dispose of it. This is
        # done to trigger the control's Paint event and force it to redraw.
        # Since in toga, Hwnds could be created at inappropriate times.
        # As an example, without this fix, running the OptionContainer
        # example app should give an error, like:
        #
        # System.ArgumentOutOfRangeException: InvalidArgument=Value of '0' is not valid
        # for 'index'.
        # Parameter name: index
        #    at System.Windows.Forms.TabControl.GetTabPage(Int32 index)
        self.native.CreateGraphics().Dispose()

    @abstractmethod
    def create(self): ...

    def set_app(self, app):
        # No special handling required
        pass

    def set_window(self, window):
        self.scale_font()

    @property
    def dpi_scale(self):
        window = self.interface.window
        if window:
            return window._impl.dpi_scale
        else:
            return 1

    @property
    def container(self):
        return self._container

    @container.setter
    def container(self, container):
        if self._container:
            self._container.remove_content(self)

        self._container = container
        if container:
            container.add_content(self)

        for child in self.interface.children:
            child._impl.container = container

        self.refresh()

        # Background color needs to be reapplied on widget parent change as WinForms
        # doesn't actually support transparency. It just copies the parent's
        # BackColor to the widget. So, if a widget's parent changes then we need
        # to reapply background_color to copy the new parent's BackColor.
        self.set_background_color(self.interface.style.background_color)

    def get_tab_index(self):
        return self.native.TabIndex

    def set_tab_index(self, tab_index):
        self.native.TabIndex = tab_index

    def get_enabled(self):
        return self.native.Enabled

    def set_enabled(self, value):
        self.native.Enabled = value

    def focus(self):
        self.native.Focus()

    # APPLICATOR

    def set_bounds(self, x, y, width, height):
        self.native.Size = Size(*map(self.scale_in, (width, height)))
        self.native.Location = Point(*map(self.scale_in, (x, y)))

    def set_text_align(self, alignment):
        # By default, text alignment can't be changed
        pass

    def set_hidden(self, hidden):
        self.native.Visible = not hidden

    def set_font(self, font):
        self.original_font = font._impl.native
        self.scale_font()

    def scale_font(self):
        font = self.original_font
        window = self.interface.window
        if window:
            font = window._impl.scale_font(self.original_font)
        self.native.Font = font

    def set_color(self, color):
        if color is None:
            self.native.ForeColor = SystemColors.WindowText
        else:
            self.native.ForeColor = native_color(color)

    def set_background_color(self, color):
        if self.interface.parent:
            parent_color = toga_color(self.interface.parent._impl.native.BackColor).rgba
        else:
            parent_color = toga_color(SystemColors.Control).rgba

        if color is None:
            if self._default_background_color is TRANSPARENT:
                requested_color = rgba(0, 0, 0, 0)
            else:
                requested_color = self._default_background_color.rgba
        elif color is TRANSPARENT:
            requested_color = rgba(0, 0, 0, 0)
        else:
            requested_color = color.rgba

        blended_color = requested_color.blend_over(parent_color)
        self.native.BackColor = native_color(blended_color)

        for child in self.interface.children:
            child._impl.set_background_color(child.style.background_color)

    # INTERFACE

    def add_child(self, child):
        child.container = self.container

    def insert_child(self, index, child):
        self.add_child(child)

    def remove_child(self, child):
        child.container = None

    def refresh(self):
        # Default values; may be overwritten by rehint().
        self.interface.intrinsic.width = at_least(self.interface._MIN_WIDTH)
        self.interface.intrinsic.height = at_least(self.interface._MIN_HEIGHT)
        self.rehint()

    def rehint(self):
        pass



import asyncio
import sys
import threading
import traceback
from asyncio import events

import System.Windows.Forms as WinForms
from System import Action
from System.Threading.Tasks import Task

from .wrapper import WeakrefCallable


class WinformsProactorEventLoop(asyncio.ProactorEventLoop):
    def run_forever(self, app):
        """Set up the asyncio event loop, integrate it with the Winforms event loop, and
        start the application.

        This largely duplicates the setup behavior of the default Proactor
        run_forever implementation.

        :param app_context: The WinForms.ApplicationContext instance
            controlling the lifecycle of the app.
        """
        # Python 3.8 added an implementation of run_forever() in
        # ProactorEventLoop. The only part that actually matters is the
        # refactoring that moved the initial call to stage _loop_self_reading;
        # it now needs to be created as part of run_forever; otherwise the
        # event loop locks up, because there won't be anything for the
        # select call to process.
        self.call_soon(self._loop_self_reading)

        # Remember the application.
        self.app = app

        # Set up the Proactor.
        if sys.version_info < (3, 13):
            # The code between the following markers should be exactly the same
            # as the official CPython implementation, up to the start of the
            # `while True:` part of run_forever() (see
            # BaseEventLoop.run_forever() in Lib/ascynio/base_events.py). In
            # Python 3.13.0a2, this was refactored into the
            # `_run_forever_setup()` helper. We run testbed on Py3.10, so the
            # else branch is marked nocover.
            # === START BaseEventLoop.run_forever() setup ===
            self._check_closed()
            if self.is_running():  # pragma: no cover
                raise RuntimeError("This event loop is already running")
            if events._get_running_loop() is not None:  # pragma: no cover
                raise RuntimeError(
                    "Cannot run the event loop while another loop is running"
                )
            self._set_coroutine_origin_tracking(self._debug)
            self._thread_id = threading.get_ident()
            self._old_agen_hooks = sys.get_asyncgen_hooks()
            sys.set_asyncgen_hooks(
                firstiter=self._asyncgen_firstiter_hook,
                finalizer=self._asyncgen_finalizer_hook,
            )

            events._set_running_loop(self)
            # === END BaseEventLoop.run_forever() setup ===
        else:  # pragma: no cover
            self._orig_state = self._run_forever_setup()

        # Rather than going into a `while True:` loop, we're going to use the
        # Winforms event loop to queue a tick() message that will cause a
        # single iteration of the asyncio event loop to be executed. Each time
        # we do this, we queue *another* tick() message in 5ms time. In this
        # way, we'll get a continuous stream of tick() calls, without blocking
        # the Winforms event loop. We also add a handler for ApplicationExit
        # to ensure that loop cleanup occurs when the app exits.

        WinForms.Application.ApplicationExit += WeakrefCallable(
            self.winforms_application_exit
        )

        # Queue the first asyncio tick.
        self.enqueue_tick()

        # Start the Winforms event loop.
        self._inner_loop = None
        WinForms.Application.Run(self.app.app_context)

    def enqueue_tick(self):
        # Queue a call to tick in 5ms.
        self.task = Action[Task](self.tick)
        Task.Delay(5).ContinueWith(self.task)

    # This function doesn't report as covered because it runs on a
    # non-Python-created thread (see App.run_app). But it must actually be
    # covered, otherwise nothing would work.
    def tick(self, *args, **kwargs):  # pragma: no cover
        """Cause a single iteration of the event loop to run on the main GUI thread."""
        action = Action(self.run_once_recurring)
        self.app.app_dispatcher.Invoke(action)

    # The native dialog `Show` methods are all blocking, as they run an inner native
    # event loop. Call them via this method to ensure the inner loop is correctly linked
    # with this Python loop.
    def start_inner_loop(self, callback, *args):
        assert self._inner_loop is None
        self._inner_loop = (callback, args)

    # We can't get coverage for app shutdown, so this handler must be no-cover.
    def winforms_application_exit(self, app, event):  # pragma: no cover
        """Perform cleanup that needs to occur when the app exits.

        This largely duplicates the "finally" behavior of the default Proactor
        run_forever implementation.
        """
        if sys.version_info < (3, 13):
            # If we're stopping, we can do the "finally" handling from
            # the BaseEventLoop run_forever(). In Python 3.13.0a2, this
            # was refactored into the `_run_forever_cleanup()` helper.
            # We run testbed on Py3.12, so the else branch is marked
            # nocover.
            # === START BaseEventLoop.run_forever() finally handling ===
            self._stopping = False
            self._thread_id = None
            events._set_running_loop(None)
            self._set_coroutine_origin_tracking(False)
            sys.set_asyncgen_hooks(*self._old_agen_hooks)
            # === END BaseEventLoop.run_forever() finally handling ===
        else:  # pragma: no cover
            self._run_forever_cleanup()

    def run_once_recurring(self):
        """Run one iteration of the event loop, and enqueue the next iteration (if we're
        not stopping).
        """
        try:
            # If the app is exiting, stop the asyncio event loop.
            # Otherwise, perform one more tick of the event loop.
            # We can't get coverage of app shutdown, so that branch
            # is marked no cover
            if self.app._is_exiting:
                self.stop()  # pragma: no cover
            else:
                self._run_once()

            # Enqueue the next tick, and make sure there will be *something*
            # to be processed. If you don't ensure there is at least one
            # message on the queue, the select() call will block, locking
            # the app.
            self.enqueue_tick()
            self.call_soon(self._loop_self_reading)

            if self._inner_loop:
                callback, args = self._inner_loop
                self._inner_loop = None
                callback(*args)

        # Exceptions thrown by this method will be silently ignored.
        except BaseException:  # pragma: no cover
            traceback.print_exc()



from .table import TableProbe


class DetailedListProbe(TableProbe):
    supports_actions = False
    supports_refresh = False

    def assert_cell_content(self, row, title, subtitle, icon=None):
        super().assert_cell_content(row, 0, title, icon=icon)
        super().assert_cell_content(row, 1, subtitle)



from ctypes import c_void_p, windll, wintypes

from System import Environment

user32 = windll.user32


# https://learn.microsoft.com/en-us/windows/win32/hidpi/dpi-awareness-context
DPI_AWARENESS_CONTEXT_UNAWARE = -1
DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2 = -4

# https://www.lifewire.com/windows-version-numbers-2625171
win_version = Environment.OSVersion.Version
if (win_version.Major, win_version.Minor, win_version.Build) >= (10, 0, 15063):
    SetProcessDpiAwarenessContext = user32.SetProcessDpiAwarenessContext
    SetProcessDpiAwarenessContext.restype = wintypes.BOOL
    SetProcessDpiAwarenessContext.argtypes = [c_void_p]

    SetThreadDpiAwarenessContext = user32.SetThreadDpiAwarenessContext
    SetThreadDpiAwarenessContext.restype = c_void_p
    SetThreadDpiAwarenessContext.argtypes = [c_void_p]

else:  # pragma: no cover
    print(
        "WARNING: Your Windows version doesn't support DPI Awareness setting. "
        "We recommend you upgrade to at least Windows 10 version 1703."
    )
    SetProcessDpiAwarenessContext = SetThreadDpiAwarenessContext = None


# https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-monitorfromrect
MONITOR_DEFAULTTONEAREST = 2

MonitorFromRect = user32.MonitorFromRect
MonitorFromRect.restype = wintypes.HMONITOR
MonitorFromRect.argtypes = [wintypes.LPRECT, wintypes.DWORD]



from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from travertino.size import at_least

from .base import Widget


class Divider(Widget):
    def create(self):
        self.native = WinForms.Label()
        self.native.BorderStyle = WinForms.BorderStyle.Fixed3D
        self.native.AutoSize = False

        self._direction = self.interface.HORIZONTAL

    def set_background_color(self, value):
        # Do nothing, since background color of Divider shouldn't be changed.
        pass

    def get_direction(self):
        return self._direction

    def set_direction(self, value):
        self._direction = value
        if value == self.interface.HORIZONTAL:
            self.native.Height = 2
            self.native.Width = 0
        else:
            self.native.Height = 0
            self.native.Width = 2

    def rehint(self):
        if self.get_direction() == self.interface.HORIZONTAL:
            self.interface.intrinsic.width = self.scale_out(
                at_least(self.native.Width), ROUND_UP
            )
            self.interface.intrinsic.height = self.scale_out(
                self.native.Height, ROUND_UP
            )
        else:
            self.interface.intrinsic.width = self.scale_out(self.native.Width, ROUND_UP)
            self.interface.intrinsic.height = self.scale_out(
                at_least(self.native.Height), ROUND_UP
            )



from datetime import time

from System import ArgumentOutOfRangeException
from System.Windows.Forms import DateTimePickerFormat

from .dateinput import DateTimeInputProbe


class TimeInputProbe(DateTimeInputProbe):
    supports_seconds = True

    def __init__(self, widget):
        super().__init__(widget)
        assert self.native.Format == DateTimePickerFormat.Time
        assert self.native.ShowUpDown

    def py_value(self, native_value):
        assert (native_value.Year, native_value.Month, native_value.Day) == (1970, 1, 1)
        return time(native_value.Hour, native_value.Minute, native_value.Second)

    async def change(self, delta):
        try:
            self.native.Value = self.native.Value.AddMinutes(delta)
        except ArgumentOutOfRangeException:
            pass
        await self.redraw(f"Change value by {delta} minutes")



from System.Windows.Forms import Panel, SplitContainer

from .base import SimpleProbe


class SplitContainerProbe(SimpleProbe):
    native_class = SplitContainer
    border_size = 2
    direction_change_preserves_position = True

    def __init__(self, widget):
        super().__init__(widget)

        for panel in [self.native.Panel1, self.native.Panel2]:
            assert panel.Controls.Count == 1
            assert isinstance(panel.Controls[0], Panel)

    def move_split(self, position):
        self.native.SplitterDistance = round(position * self.scale_factor)

    async def wait_for_split(self):
        pass



from toga import NotImplementedWarning

from . import dialogs
from .app import App
from .command import Command
from .fonts import Font
from .icons import Icon
from .images import Image
from .paths import Paths
from .statusicons import MenuStatusIcon, SimpleStatusIcon, StatusIconSet
from .widgets.box import Box
from .widgets.button import Button
from .widgets.canvas import Canvas
from .widgets.dateinput import DateInput
from .widgets.detailedlist import DetailedList
from .widgets.divider import Divider
from .widgets.imageview import ImageView
from .widgets.label import Label
from .widgets.mapview import MapView
from .widgets.multilinetextinput import MultilineTextInput
from .widgets.numberinput import NumberInput
from .widgets.optioncontainer import OptionContainer
from .widgets.passwordinput import PasswordInput
from .widgets.progressbar import ProgressBar
from .widgets.scrollcontainer import ScrollContainer
from .widgets.selection import Selection
from .widgets.slider import Slider
from .widgets.splitcontainer import SplitContainer
from .widgets.switch import Switch
from .widgets.table import Table
from .widgets.textinput import TextInput
from .widgets.timeinput import TimeInput
from .widgets.webview import WebView
from .window import MainWindow, Window


def not_implemented(feature):
    NotImplementedWarning.warn("Winforms", feature)  # pragma: nocover


__all__ = [
    "not_implemented",
    "App",
    "Command",
    # Resources
    "Font",
    "Icon",
    "Image",
    "Paths",
    "dialogs",
    # Status Icons
    "MenuStatusIcon",
    "SimpleStatusIcon",
    "StatusIconSet",
    # Widgets
    "Box",
    "Button",
    "Canvas",
    "DateInput",
    "DetailedList",
    "Divider",
    "ImageView",
    "Label",
    "MapView",
    "MultilineTextInput",
    "NumberInput",
    "OptionContainer",
    "PasswordInput",
    "ProgressBar",
    "ScrollContainer",
    "Selection",
    "Slider",
    "SplitContainer",
    "Switch",
    "Table",
    "TextInput",
    "TimeInput",
    "WebView",
    # Windows
    "Window",
    "MainWindow",
]


def __getattr__(name):  # pragma: no cover
    raise NotImplementedError(f"Toga's Winforms backend doesn't implement {name}")



import System.Windows.Forms as WinForms
from System.Drawing import SystemColors
from travertino.size import at_least

from toga_winforms.colors import native_color
from toga_winforms.libs.fonts import HorizontalTextAlignment

from ..libs.wrapper import WeakrefCallable
from .textinput import TextInput


class MultilineTextInput(TextInput):
    def create(self):
        # TextBox doesn't support automatic scroll bar visibility, so we use RichTextBox
        # (https://stackoverflow.com/a/612234).
        self.native = WinForms.RichTextBox()
        self.native.Multiline = True
        self.native.TextChanged += WeakrefCallable(self.winforms_text_changed)

        # When moving focus with the tab key, the Enter/Leave event handlers see the
        # wrong value of ContainsFocus, so we use GotFocus/LostFocus instead.
        self.native.GotFocus += WeakrefCallable(self.winforms_got_focus)
        self.native.LostFocus += WeakrefCallable(self.winforms_lost_focus)

        # Dummy values used during initialization
        self._placeholder = ""
        self._placeholder_visible = True
        self.set_color(None)

    def winforms_got_focus(self, sender, event):
        # If the placeholder is visible when we gain focus, the widget is empty;
        # so make the native text empty and hide the placeholder.
        if self._placeholder_visible:
            self.native.Text = ""
            self._set_placeholder_visible(False)

    def winforms_lost_focus(self, sender, event):
        # When we lose focus, if the widget is empty, we need to show the
        # placeholder again.
        if self.native.Text == "":
            self._set_placeholder_visible(True)

    def set_placeholder(self, value):
        self._placeholder = value
        if self._placeholder_visible:
            self.native.Text = value

    def get_value(self):
        # If the placeholder is visible, we know the widget has no value
        if self._placeholder_visible:
            return ""
        return self.native.Text

    def set_value(self, value):
        # If the value is empty, the placeholder is only visible if the widget
        # does *not* currently have focus.
        if value == "" and not self.native.ContainsFocus:
            self.native.Text = value
            self._set_placeholder_visible(True)
        else:
            self._set_placeholder_visible(False)
            self.native.Text = value

    # This method is necessary to override the TextInput base class.
    def rehint(self):
        self.interface.intrinsic.width = at_least(self.interface._MIN_WIDTH)
        self.interface.intrinsic.height = at_least(self.interface._MIN_HEIGHT)

    def winforms_text_changed(self, sender, event):
        # Showing and hiding the placeholder should not cause an interface event.
        if not self._placeholder_visible:
            self.interface.on_change()

    def _set_placeholder_visible(self, visible):
        # Changing ForeColor causes a native TextChanged event, so the order of these
        # lines is important.
        if visible:
            self._placeholder_visible = True
            self.native.Text = self._placeholder
            self.native.ForeColor = SystemColors.GrayText
        elif self._placeholder_visible:
            self.native.ForeColor = self._color
            self._placeholder_visible = False

    def set_color(self, color):
        self._color = (
            SystemColors.WindowText if (color is None) else native_color(color)
        )
        if not self._placeholder_visible:
            self.native.ForeColor = self._color

    def set_text_align(self, value):
        original_selection = (self.native.SelectionStart, self.native.SelectionLength)
        self.native.SelectAll()
        self.native.SelectionAlignment = HorizontalTextAlignment(value)
        self.native.SelectionStart, self.native.SelectionLength = original_selection

    def scroll_to_bottom(self):
        self.native.SelectionStart = len(self.native.Text)
        self.native.ScrollToCaret()

    def scroll_to_top(self):
        self.native.SelectionStart = 0
        self.native.ScrollToCaret()



from System.Drawing import ContentAlignment
from System.Windows.Forms import HorizontalAlignment

from toga.colors import rgba
from toga.style.pack import BOTTOM, CENTER, LEFT, RIGHT, TOP


def toga_color(color):
    return rgba(color.R, color.G, color.B, color.A / 255)


def toga_x_text_align(alignment):
    return {
        ContentAlignment.TopLeft: LEFT,
        ContentAlignment.MiddleLeft: LEFT,
        ContentAlignment.BottomLeft: LEFT,
        ContentAlignment.TopCenter: CENTER,
        ContentAlignment.MiddleCenter: CENTER,
        ContentAlignment.BottomCenter: CENTER,
        ContentAlignment.TopRight: RIGHT,
        ContentAlignment.MiddleRight: RIGHT,
        ContentAlignment.BottomRight: RIGHT,
        #
        HorizontalAlignment.Left: LEFT,
        HorizontalAlignment.Center: CENTER,
        HorizontalAlignment.Right: RIGHT,
    }[alignment]


def toga_y_text_align(alignment):
    return {
        ContentAlignment.TopLeft: TOP,
        ContentAlignment.TopCenter: TOP,
        ContentAlignment.TopRight: TOP,
        ContentAlignment.MiddleLeft: CENTER,
        ContentAlignment.MiddleCenter: CENTER,
        ContentAlignment.MiddleRight: CENTER,
        ContentAlignment.BottomLeft: BOTTOM,
        ContentAlignment.BottomCenter: BOTTOM,
        ContentAlignment.BottomRight: BOTTOM,
    }[alignment]



import System.Windows.Forms as WinForms

from toga.colors import TRANSPARENT

from .base import Widget


class Box(Widget):
    def create(self):
        self.native = WinForms.Panel()
        self._default_background_color = TRANSPARENT



from pathlib import Path

import PIL.Image
import pytest
from System.Drawing import Icon as WinIcon

import toga
import toga_winforms

from .probe import BaseProbe


class IconProbe(BaseProbe):
    alternate_resource = "resources/icons/blue"

    def __init__(self, app, icon):
        super().__init__()
        self.app = app
        self.icon = icon
        assert isinstance(self.icon._impl.native, WinIcon)

    def assert_icon_content(self, path):
        if path == "resources/icons/green":
            assert (
                self.icon._impl.path == self.app.paths.app / "resources/icons/green.ico"
            )
        elif path == "resources/icons/blue":
            assert (
                self.icon._impl.path == self.app.paths.app / "resources/icons/blue.png"
            )
        else:
            pytest.fail("Unknown icon resource")

    def assert_default_icon_content(self):
        assert (
            self.icon._impl.path
            == Path(toga_winforms.__file__).parent / "resources/toga.ico"
        )

    def assert_platform_icon_content(self):
        assert self.icon._impl.path == self.app.paths.app / "resources/logo-windows.ico"

    def assert_app_icon_content(self):
        # We have no real way to check we've got the right icon; use pixel peeping as a
        # guess. Construct a PIL image from the current icon.
        img = toga.Image(self.icon._impl.bitmap).as_format(PIL.Image.Image)

        # The default icon is transparent background, and brown in the center.
        assert img.getpixel((5, 5))[3] == 0
        mid_color = img.getpixel((img.size[0] // 2, img.size[1] // 2))
        assert mid_color == (130, 100, 57, 255)



import System.Windows.Forms as WinForms
from System.Drawing import Size

from .widgets.base import Scalable


class Container(Scalable):
    def __init__(self, native_parent):
        self.native_parent = native_parent
        self.native_width = self.native_height = 0
        self.content = None

        self.native_content = WinForms.Panel()
        native_parent.Controls.Add(self.native_content)

        # See comment in Widget.__init__.
        self.native_content.CreateGraphics().Dispose()

    @property
    def dpi_scale(self):
        window = self.content.interface.window if self.content else None
        if window:
            return window._impl.dpi_scale
        else:
            return 1

    @property
    def width(self):
        return self.scale_out(self.native_width)

    @property
    def height(self):
        return self.scale_out(self.native_height)

    def set_content(self, widget):
        self.clear_content()
        if widget:
            widget.container = self
            self.content = widget

    def clear_content(self):
        if self.content:
            self.content.container = None
            self.content = None

    def resize_content(self, width, height, *, force_refresh=False):
        if (self.native_width, self.native_height) != (width, height):
            self.native_width, self.native_height = (width, height)
            force_refresh = True

        if force_refresh and self.content:
            self.content.interface.refresh()

    def refreshed(self):
        layout = self.content.interface.layout
        self.apply_layout(layout.width, layout.height)

    def apply_layout(self, layout_width, layout_height):
        self.native_content.Size = Size(
            self.scale_in(max(self.width, layout_width)),
            self.scale_in(max(self.height, layout_height)),
        )

    def add_content(self, widget):
        # The default is to add new controls to the back of the Z-order.
        self.native_content.Controls.Add(widget.native)
        widget.native.BringToFront()

    def remove_content(self, widget):
        self.native_content.Controls.Remove(widget.native)



import clr
import travertino

from .libs.user32 import (
    DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2,
    SetProcessDpiAwarenessContext,
)

# Add a reference to the Winforms assembly
clr.AddReference("System.Windows.Forms")

# Add a reference to the WindowsBase assembly. This is needed to access
# System.Windows.Threading.Dispatcher.
#
# This assembly isn't exposed as a simple dot-path name; we have to extract it from the
# Global Assembly Cache (GAC). The version number and public key doesn't appear to
# change with Windows version or the underlying .NET, and has been available since
# Windows 7.
clr.AddReference(
    "WindowsBase, Version=4.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"
)


# Enable DPI awareness. This must be done before calling any other UI-related code
# (https://learn.microsoft.com/en-us/dotnet/desktop/winforms/high-dpi-support-in-windows-forms).
import System.Windows.Forms as WinForms  # noqa: E402

WinForms.Application.EnableVisualStyles()
WinForms.Application.SetCompatibleTextRenderingDefault(False)

if SetProcessDpiAwarenessContext is not None:
    if not SetProcessDpiAwarenessContext(
        DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2
    ):  # pragma: no cover
        print("WARNING: Failed to set the DPI Awareness mode for the app.")

__version__ = travertino._package_version(__file__, __name__)



import datetime
from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from System import DateTime as WinDateTime

from ..libs.wrapper import WeakrefCallable
from .base import Widget


def py_date(native_date):
    return datetime.date(native_date.Year, native_date.Month, native_date.Day)


def native_date(py_date):
    return WinDateTime(py_date.year, py_date.month, py_date.day)


class DateInput(Widget):

    def create(self):
        self.native = WinForms.DateTimePicker()
        self.native.ValueChanged += WeakrefCallable(self.winforms_value_changed)

    def get_value(self):
        return py_date(self.native.Value)

    def set_value(self, value):
        self.native.Value = native_date(value)

    def get_min_date(self):
        return py_date(self.native.MinDate)

    def set_min_date(self, value):
        self.native.MinDate = native_date(value)

    def get_max_date(self):
        return py_date(self.native.MaxDate)

    def set_max_date(self, value):
        self.native.MaxDate = native_date(value)

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )

    def winforms_value_changed(self, sender, event):
        self.interface.on_change()



import ctypes
from pathlib import Path
from time import sleep

import PIL.Image
import pytest
from System import EventArgs
from System.Drawing import Bitmap, Point
from System.Windows.Forms import Application, Cursor, ToolStripSeparator

import toga
from toga_winforms.keys import toga_to_winforms_key, winforms_to_toga_key

from .dialogs import DialogsMixin
from .probe import BaseProbe


class AppProbe(BaseProbe, DialogsMixin):
    supports_key = True
    supports_key_mod3 = False
    supports_current_window_assignment = True

    def __init__(self, app):
        super().__init__()
        self.app = app
        self.main_window = app.main_window
        # The Winforms Application class is a singleton instance
        assert self.app._impl.native == Application

    @property
    def config_path(self):
        return (
            Path.home()
            / "AppData"
            / "Local"
            / "Tiberius Yak"
            / "Toga Testbed"
            / "Config"
        )

    @property
    def data_path(self):
        return Path.home() / "AppData/Local/Tiberius Yak/Toga Testbed/Data"

    @property
    def cache_path(self):
        return (
            Path.home()
            / "AppData"
            / "Local"
            / "Tiberius Yak"
            / "Toga Testbed"
            / "Cache"
        )

    @property
    def logs_path(self):
        return Path.home() / "AppData/Local/Tiberius Yak/Toga Testbed/Logs"

    @property
    def is_cursor_visible(self):
        # Despite what the documentation says, Cursor.Current never returns null in
        # Windows 10, whether the cursor is over the window or not.
        #
        # The following code is based on https://stackoverflow.com/a/12467292, but it
        # only works when the cursor is over the window.
        form = self.main_window._impl.native
        Cursor.Position = Point(
            form.Location.X + (form.Size.Width // 2),
            form.Location.Y + (form.Size.Height // 2),
        )

        # A small delay is apparently required for the new position to take effect.
        sleep(0.1)

        class POINT(ctypes.Structure):
            _fields_ = [
                ("x", ctypes.c_long),
                ("y", ctypes.c_long),
            ]

        class CURSORINFO(ctypes.Structure):
            _fields_ = [
                ("cbSize", ctypes.c_uint32),
                ("flags", ctypes.c_uint32),
                ("hCursor", ctypes.c_void_p),
                ("ptScreenPos", POINT),
            ]

        GetCursorInfo = ctypes.windll.user32.GetCursorInfo
        GetCursorInfo.argtypes = [ctypes.POINTER(CURSORINFO)]

        info = CURSORINFO()
        info.cbSize = ctypes.sizeof(info)
        if not GetCursorInfo(ctypes.byref(info)):
            raise RuntimeError("GetCursorInfo failed")

        # `flags` is 0 or 1 in local testing, but the GitHub Actions runner always
        # returns 2 ("the system is not drawing the cursor because the user is providing
        # input through touch or pen instead of the mouse"). hCursor is more reliable.
        return info.hCursor is not None

    def unhide(self):
        pytest.xfail("This platform doesn't have an app level unhide.")

    def assert_app_icon(self, icon):
        for window in self.app.windows:
            # We have no real way to check we've got the right icon; use pixel peeping
            # as a guess. Construct a PIL image from the current icon.
            img = toga.Image(
                Bitmap.FromHicon(window._impl.native.Icon.Handle)
            ).as_format(PIL.Image.Image)

            if icon:
                # The explicit alt icon has blue background, with green at a point 1/3
                # into the image
                assert img.getpixel((5, 5)) == (211, 230, 245, 255)
                mid_color = img.getpixel((img.size[0] // 3, img.size[1] // 3))
                assert mid_color == (0, 204, 9, 255)
            else:
                # The default icon is transparent background, and brown in the center.
                assert img.getpixel((5, 5))[3] == 0
                mid_color = img.getpixel((img.size[0] // 2, img.size[1] // 2))
                assert mid_color == (130, 100, 57, 255)

    def _menu_item(self, path):
        item = self.main_window._impl.native.MainMenuStrip
        for i, label in enumerate(path):
            children = getattr(item, "Items" if i == 0 else "DropDownItems")
            child_labels = [child.Text for child in children]
            try:
                child_index = child_labels.index(label)
            except ValueError:
                raise AssertionError(
                    f"no item named {path[:i+1]}; options are {child_labels}"
                ) from None
            item = children[child_index]

        return item

    def _activate_menu_item(self, path):
        self._menu_item(path).OnClick(EventArgs.Empty)

    def activate_menu_hide(self):
        pytest.xfail("This platform doesn't present a app level hide option in menu.")

    def activate_menu_exit(self):
        self._activate_menu_item(["File", "Exit"])

    def activate_menu_about(self):
        self._activate_menu_item(["Help", "About Toga Testbed"])

    async def close_about_dialog(self):
        await self.type_character("\n")

    def activate_menu_visit_homepage(self):
        self._activate_menu_item(["Help", "Visit homepage"])

    def assert_dialog_in_focus(self, dialog):
        active_window_handle = ctypes.windll.user32.GetForegroundWindow()
        # The window class name for dialog boxes is "#32770":
        # https://learn.microsoft.com/en-us/windows/win32/winauto/dialog-box
        expected_dialog_handle = ctypes.windll.user32.FindWindowW(
            "#32770", dialog._impl.title
        )
        assert (
            expected_dialog_handle == active_window_handle
        ), "The dialog is not in focus"

    def assert_menu_item(self, path, *, enabled=True):
        item = self._menu_item(path)
        assert item.Enabled == enabled

        # Check some special cases of menu shortcuts
        try:
            shortcut = {
                ("Other", "Full command"): "Ctrl+1",
                ("Other", "Submenu1", "Disabled"): None,
                ("Commands", "No Tooltip"): "Ctrl+Down",
                ("Commands", "Sectioned"): "Ctrl+Space",
            }[tuple(path)]
        except KeyError:
            pass
        else:
            assert item.ShortcutKeyDisplayString == shortcut

    def assert_menu_order(self, path, expected):
        menu = self._menu_item(path)

        assert len(menu.DropDownItems) == len(expected)
        for item, title in zip(menu.DropDownItems, expected):
            if title == "---":
                assert isinstance(item, ToolStripSeparator)
            else:
                assert item.Text == title

    def assert_system_menus(self):
        self.assert_menu_item(["File", "New Example Document"], enabled=True)
        self.assert_menu_item(["File", "New Read-only Document"], enabled=True)
        self.assert_menu_item(["File", "Open..."], enabled=True)
        self.assert_menu_item(["File", "Save"], enabled=True)
        self.assert_menu_item(["File", "Save As..."], enabled=True)
        self.assert_menu_item(["File", "Save All"], enabled=True)
        self.assert_menu_item(["File", "Preferences"], enabled=False)
        self.assert_menu_item(["File", "Exit"])

        self.assert_menu_item(["Help", "Visit homepage"])
        self.assert_menu_item(["Help", "About Toga Testbed"])

    def activate_menu_close_window(self):
        pytest.xfail("This platform doesn't have a window management menu")

    def activate_menu_close_all_windows(self):
        pytest.xfail("This platform doesn't have a window management menu")

    def activate_menu_minimize(self):
        pytest.xfail("This platform doesn't have a window management menu")

    def keystroke(self, combination):
        return winforms_to_toga_key(toga_to_winforms_key(combination))

    async def restore_standard_app(self):
        # No special handling needed to restore standard app.
        await self.redraw("Restore to standard app")

    async def open_initial_document(self, monkeypatch, document_path):
        pytest.xfail("Winforms doesn't require initial document support")

    def open_document_by_drag(self, document_path):
        pytest.xfail("Winforms doesn't support opening documents by drag")

    def has_status_icon(self, status_icon):
        return status_icon._impl.native is not None

    def status_menu_items(self, status_icon):
        if status_icon._impl.native.ContextMenu:
            return [
                {
                    "-": "---",
                    "About Toga Testbed": "**ABOUT**",
                    "Exit": "**EXIT**",
                }.get(str(item.Text), str(item.Text))
                for item in status_icon._impl.native.ContextMenu.MenuItems
            ]
        else:
            # It's a button status item
            return None

    def activate_status_icon_button(self, item_id):
        # Winforms doesn't provide an OnClick to trigger clicks, so we have to fake it
        # at the level of the impl.
        self.app.status_icons[item_id]._impl.winforms_click(
            self.app.status_icons[item_id]._impl.native,
            EventArgs.Empty,
        )

    def activate_status_menu_item(self, item_id, title):
        menu = self.app.status_icons[item_id]._impl.native.ContextMenu
        item = {item.Text: item for item in menu.MenuItems}[title]
        item.OnClick(EventArgs.Empty)



import System.Windows.Forms

from .base import SimpleProbe


class SwitchProbe(SimpleProbe):
    native_class = System.Windows.Forms.CheckBox

    @property
    def text(self):
        # Normalize the zero width space to the empty string.
        if self.native.Text == "\u200b":
            return ""
        return self.native.Text



import operator
import re
from functools import reduce
from string import ascii_lowercase

import System.Windows.Forms as WinForms

from toga.keys import Key

WINFORMS_MODIFIERS = {
    Key.MOD_1: WinForms.Keys.Control,
    Key.MOD_2: WinForms.Keys.Alt,
    Key.SHIFT: WinForms.Keys.Shift,
}

WINFORMS_KEYS = {
    Key.ESCAPE.value: WinForms.Keys.Escape,
    # Key.BACK_QUOTE.value: WinForms.Keys.Oemtilde,  # No idea what the code should be
    Key.MINUS.value: WinForms.Keys.OemMinus,
    Key.EQUAL.value: WinForms.Keys.Oemplus,
    Key.CAPSLOCK.value: WinForms.Keys.CapsLock,
    Key.TAB.value: WinForms.Keys.Tab,
    Key.OPEN_BRACKET.value: WinForms.Keys.OemOpenBrackets,
    Key.CLOSE_BRACKET.value: WinForms.Keys.OemCloseBrackets,
    Key.BACKSLASH.value: WinForms.Keys.OemQuotes,  # NFI what is going on here
    Key.SEMICOLON.value: WinForms.Keys.OemSemicolon,
    Key.QUOTE.value: WinForms.Keys.Oemtilde,  # NFI what is going on here
    Key.COMMA.value: WinForms.Keys.Oemcomma,
    Key.FULL_STOP.value: WinForms.Keys.OemPeriod,
    Key.SLASH.value: WinForms.Keys.OemQuestion,  # Key uses the shifted form
    Key.SPACE.value: WinForms.Keys.Space,
    Key.PAGE_UP.value: WinForms.Keys.PageUp,
    Key.PAGE_DOWN.value: WinForms.Keys.PageDown,
    Key.INSERT.value: WinForms.Keys.Insert,
    Key.DELETE.value: WinForms.Keys.Delete,
    Key.HOME.value: WinForms.Keys.Home,
    Key.END.value: WinForms.Keys.End,
    Key.UP.value: WinForms.Keys.Up,
    Key.DOWN.value: WinForms.Keys.Down,
    Key.LEFT.value: WinForms.Keys.Left,
    Key.RIGHT.value: WinForms.Keys.Right,
    Key.NUMLOCK.value: WinForms.Keys.NumLock,
    Key.NUMPAD_DECIMAL_POINT.value: WinForms.Keys.Decimal,
    Key.SCROLLLOCK.value: WinForms.Keys.Scroll,
    Key.MENU.value: WinForms.Keys.Menu,
}
WINFORMS_KEYS.update(
    {str(digit): getattr(WinForms.Keys, f"D{digit}") for digit in range(10)}
)
WINFORMS_KEYS.update(
    {
        getattr(Key, f"NUMPAD_{digit}").value: getattr(WinForms.Keys, f"NumPad{digit}")
        for digit in range(10)
    }
)

SHIFTED_KEYS = {symbol: number for symbol, number in zip("!@#$%^&*()", "1234567890")}
SHIFTED_KEYS.update(
    {lower.upper(): lower for lower in ascii_lowercase},
)
SHIFTED_KEYS.update(
    {
        "~": "`",
        "_": "-",
        "+": "=",
        "{": "[",
        "}": "]",
        "|": "\\",
        ":": ";",
        '"': "'",
        "<": ",",
        ">": ".",
        "?": "/",
    }
)


def toga_to_winforms_key(key):
    # Convert a Key object into string form.
    try:
        key = key.value
    except AttributeError:
        pass

    codes = []
    for modifier, modifier_code in WINFORMS_MODIFIERS.items():
        if modifier.value in key:
            codes.append(modifier_code)
            key = key.replace(modifier.value, "")

    if lower := SHIFTED_KEYS.get(key):
        key = lower
        codes.append(WinForms.Keys.Shift)

    try:
        codes.append(WINFORMS_KEYS[key])
    except KeyError:
        if match := re.fullmatch(r"<(.+)>", key):
            key = match[1]
        try:
            codes.append(getattr(WinForms.Keys, key.title()))
        except AttributeError:  # pragma: no cover
            raise ValueError(f"unknown key: {key!r}") from None

    return reduce(operator.or_, codes)


def toga_to_winforms_shortcut(key):
    # The Winforms key enum is... daft. The "oem" key values render as "Oem" or
    # "Oemcomma", so we need to *manually* set the display text for the key
    # shortcut.

    # Convert a Key object into string form.
    try:
        key = key.value
    except AttributeError:
        key = key

    # Replace modifiers with the Winforms text
    display = []
    for toga_keyval, winforms_keyval in [
        (Key.MOD_1.value, "Ctrl"),
        (Key.MOD_2.value, "Alt"),
        (Key.SHIFT.value, "Shift"),
    ]:
        if toga_keyval in key:
            display.append(winforms_keyval)
            key = key.replace(toga_keyval, "")

    if key == " ":
        display.append("Space")
    else:
        # Convert non-printable characters to printable
        if match := re.fullmatch(r"<(.+)>", key):
            key = match[1]

        # All remaining text is displayed in title case. Shift will already be
        # in the shortcut if it's an upper case letter; it's title() rather
        # than upper() because we want both a->A and esc -> Esc
        display.append(key.title())

    return "+".join(display)


def winforms_to_toga_key(code):
    modifiers = set()

    code_names = str(code).split(", ")
    for toga_mod, code in WINFORMS_MODIFIERS.items():
        try:
            code_names.remove(str(code))
        except ValueError:
            pass
        else:
            modifiers.add(toga_mod)

    assert len(code_names) == 1
    for toga_value, code in WINFORMS_KEYS.items():
        if str(code) == code_names[0]:
            break
    else:
        toga_value = code_names[0].lower()
        if len(toga_value) > 1:
            toga_value = f"<{toga_value}>"

    if (Key.SHIFT in modifiers) and (toga_value not in ascii_lowercase):
        for symbol, number in SHIFTED_KEYS.items():
            if toga_value == number:
                toga_value = symbol
                modifiers.remove(Key.SHIFT)

    return {"key": Key(toga_value), "modifiers": modifiers}



from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from travertino.size import at_least

from toga.colors import TRANSPARENT
from toga_winforms.libs.fonts import TextAlignment

from .base import Widget


class Label(Widget):
    def create(self):
        self.native = WinForms.Label()
        self._default_background_color = TRANSPARENT
        self.native.AutoSizeMode = WinForms.AutoSizeMode.GrowAndShrink

    def set_text_align(self, value):
        self.native.TextAlign = TextAlignment(value)

    def get_text(self):
        return self.native.Text

    def set_text(self, value):
        self.native.Text = value

    def rehint(self):
        self.interface.intrinsic.width = self.scale_out(
            at_least(self.native.PreferredSize.Width), ROUND_UP
        )
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



from decimal import ROUND_UP

import System.Windows.Forms as WinForms

from .base import Widget

# Implementation notes
# ====================
#
# * ProgressBar uses an integer to track for progress. Toga's values are
#   floats; so we multiply the Toga value by 1000 so that we get 3dp of
#   functional progress fidelity.
#
# * There's no native way to identify an indeterminate progress bar. We can't
#   use a max value of 0 as a marker, because a max value of 0 causes the
#   animation to stop. Therefore, we track indeterminate status independently.
#
# * Winforms progress bars are *always* animating; there's no "non-running"
#   state. Therefore, we track running status independently.


class ProgressBar(Widget):
    TOGA_SCALE = 1000

    def create(self):
        self.native = WinForms.ProgressBar()
        self.set_stopping_style()

        self._running = False
        self._determinate = True

    def is_running(self):
        return self._running

    def start(self):
        self._running = True
        self.set_running_style()

    def stop(self):
        self._running = False
        self.set_stopping_style()

    def get_max(self):
        if not self._determinate:
            return None
        return float(self.native.Maximum / self.TOGA_SCALE)

    def set_max(self, value):
        if value is None:
            self.native.Maximum = 1000
            self.native.Value = 0
            self._determinate = False
        else:
            self.native.Minimum = 0
            self.native.Maximum = int(value * self.TOGA_SCALE)
            self._determinate = True

        if self.is_running():
            self.set_running_style()
        else:
            self.set_stopping_style()

    def set_running_style(self):
        if not self._determinate and self._running:
            self.native.Style = WinForms.ProgressBarStyle.Marquee
        else:
            self.native.Style = WinForms.ProgressBarStyle.Continuous

    def set_stopping_style(self):
        self.native.Style = WinForms.ProgressBarStyle.Continuous

    def get_value(self):
        if not self._determinate:
            return None

        return float(self.native.Value / self.TOGA_SCALE)

    def set_value(self, value):
        self.native.Value = int(value * self.TOGA_SCALE)

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



from pathlib import Path

from toga import App


class Paths:
    def __init__(self, interface):
        self.interface = interface

    @property
    def author(self):
        # No coverage testing of this because we can't easily configure
        # the app to have no author.
        if App.app.author is None:  # pragma: no cover
            return "Unknown"
        return App.app.author

    def get_config_path(self):
        return (
            Path.home()
            / "AppData"
            / "Local"
            / self.author
            / App.app.formal_name
            / "Config"
        )

    def get_data_path(self):
        return (
            Path.home()
            / "AppData"
            / "Local"
            / self.author
            / App.app.formal_name
            / "Data"
        )

    def get_cache_path(self):
        return (
            Path.home()
            / "AppData"
            / "Local"
            / self.author
            / App.app.formal_name
            / "Cache"
        )

    def get_logs_path(self):
        return (
            Path.home()
            / "AppData"
            / "Local"
            / self.author
            / App.app.formal_name
            / "Logs"
        )



from decimal import ROUND_DOWN

from System.Drawing import Point
from System.Windows.Forms import Panel, SystemInformation
from travertino.node import Node

from toga_winforms.container import Container

from ..libs.wrapper import WeakrefCallable
from .base import Widget

# On Windows, scroll bars usually appear only when the content is larger than the
# container. However, this complicates layout. For example, if the content fits in
# the container horizontally but is taller vertically, we then need to do a second
# layout pass with a slightly narrower horizontal size to account for the vertical
# scroll bar (https://stackoverflow.com/questions/28418026).
#
# A previous attempt to avoid this, by making the scroll bars always visible, was not
# successful (see commit "Initial Winforms implementation" on 2023-07-10). Although the
# ScrollableControl API does provide Visible and Enabled properties for each scroll bar,
# they behave in confusing and undocumented ways, e.g.:
#
#  * https://stackoverflow.com/questions/8690643
#  * https://stackoverflow.com/questions/5489273
#
# So the current implementation just uses the default AutoScroll behavior, and does a
# second layout pass where necessary.


class ScrollContainer(Widget, Container):
    def create(self):
        self.native = Panel()
        self.native.AutoScroll = True
        Container.__init__(self, self.native)

        # The Scroll event only fires on direct interaction with the scroll bar. It
        # doesn't fire when using the mouse wheel, and it doesn't fire when setting
        # AutoScrollPosition either, despite the documentation saying otherwise.
        self.native.Scroll += WeakrefCallable(self.winforms_scroll)
        self.native.MouseWheel += WeakrefCallable(self.winforms_scroll)

    def winforms_scroll(self, sender, event):
        self.interface.on_scroll()

    def set_bounds(self, x, y, width, height):
        super().set_bounds(x, y, width, height)
        self.resize_content(
            self.scale_in(width, ROUND_DOWN),
            self.scale_in(height, ROUND_DOWN),
        )

    def refreshed(self):
        full_width, full_height = (self.native_width, self.native_height)
        inset_width = full_width - SystemInformation.VerticalScrollBarWidth
        inset_height = full_height - SystemInformation.HorizontalScrollBarHeight
        layout = self.interface.content.layout

        # Temporarily reduce container size to account for scroll bars (see explanation
        # at the top of this file).
        def apply_insets():
            need_scrollbar = False
            if self.vertical and (layout.height > self.height):
                need_scrollbar = True
                self.native_width = inset_width
            if self.horizontal and (layout.width > self.width):
                need_scrollbar = True
                self.native_height = inset_height
            return need_scrollbar

        if apply_insets():
            # Bypass Widget.refresh to avoid a recursive call to `refreshed`.
            Node.refresh(self.interface.content, self)

            # In borderline cases, adding one scroll bar may cause the other one to be
            # needed as well.
            apply_insets()

        # Crop any non-scrollable dimensions to the available size.
        self.apply_layout(
            layout.width if self.horizontal else 0,
            layout.height if self.vertical else 0,
        )

        # Restore the original container size so it'll be used in the next call to
        # `refresh` or `resize_content`.
        self.native_width, self.native_height = full_width, full_height

        # Do this to prevent horizontal scroll bar from becoming permanently visible
        # on dpi scaling changes.
        self.native.AutoScroll = False
        self.native.AutoScroll = True

    def get_horizontal(self):
        return self.horizontal

    def set_horizontal(self, value):
        self.horizontal = value
        if not value:
            self.interface.on_scroll()
        if self.interface.content:
            self.interface.content.refresh()

    def get_vertical(self):
        return self.vertical

    def set_vertical(self, value):
        self.vertical = value
        if not value:
            self.interface.on_scroll()
        if self.interface.content:
            self.interface.content.refresh()

    def get_vertical_position(self):
        return self.scale_out(abs(self.native.AutoScrollPosition.Y))

    def get_horizontal_position(self):
        return self.scale_out(abs(self.native.AutoScrollPosition.X))

    def get_max_vertical_position(self):
        return self.scale_out(
            max(0, self.native_content.Height - self.native.ClientSize.Height)
        )

    def get_max_horizontal_position(self):
        return self.scale_out(
            max(0, self.native_content.Width - self.native.ClientSize.Width)
        )

    def set_position(self, horizontal_position, vertical_position):
        self.native.AutoScrollPosition = Point(
            self.scale_in(horizontal_position),
            self.scale_in(vertical_position),
        )
        self.interface.on_scroll()
